(* ============================================================ *)
(*  IlasmUtil is the module which writes ILASM file structures  *)
(*  Copyright (c) John Gough 1999, 2000.			*)
(* ============================================================ *)

MODULE IlasmUtil;

  IMPORT 
	GPCPcopyright,
	RTS,
	Console,
	GPText,
	LitValue,
	GPTextFiles,
	CompState,
	ClassMaker,
	LightningBase,
	D := Symbols,
	G := Builtin,
	Id  := IdDesc,
	Ty  := TypeDesc,
	Asm := IlasmCodes;

(* ============================================================ *)

   CONST
	(* various ILASM-specific runtime name strings *)
	libVersion* = "mscorlib { .ver 2000:14:1812:10 }";
	catchStr*   = "      catch [mscorlib]System.Exception";
	initPrefix  = "instance void ";
	initSuffix  = ".ctor() ";
	initString  = ".ctor";
        managedStr  = "il managed";
	specialStr  = "public specialname rtspecialname ";
	cctorStr    = "static void .cctor() ";
	objectInit* = "instance void [mscorlib]System.Object::.ctor() ";
	mainStr     = "public static void Main(class System.String[]) ";
	copyHead    = "public void __copy__(";
	valueType*  = "[mscorlib]System.ValueType";

	sysExit*    = "void    [mscorlib]System.Environment::Exit(int32)";
	toUpper*    = "wchar   [mscorlib]System.Char::ToUpper(wchar) ";
	dFloor*     = "float64 [mscorlib]System.Math::Floor(float64) ";
	dAbs*       = "float64 [mscorlib]System.Math::Abs(float64) ";
	fAbs*       = "float32 [mscorlib]System.Math::Abs(float32) ";
	iAbs*       = "int32   [mscorlib]System.Math::Abs(int32) ";
	lAbs*       = "int64   [mscorlib]System.Math::Abs(int64) ";

   CONST
	putArgStr  = "class System.String[] [RTS]ProgArgs::argList";
	vStr2ChO*  = "wchar[] [RTS]CP_rts::strToChO(class System.String)";
	vStr2ChF*  = "void [RTS]CP_rts::StrToChF(wchar[], class System.String)";

   CONST (* Runtime support for CP's MOD,DIV operations *)
	CpModI*     = "int32 [RTS]CP_rts::CpModI(int32, int32)";
	CpDivI*     = "int32 [RTS]CP_rts::CpDivI(int32, int32)";
	CpModL*     = "int64 [RTS]CP_rts::CpModL(int64, int64)";
	CpDivL*     = "int64 [RTS]CP_rts::CpDivL(int64, int64)";

	(* various ILASM-specific runtime name strings *)
	aStrLen*    = "int32 [RTS]CP_rts::chrArrLength(wchar[])";
	aStrChk*    = "void  [RTS]CP_rts::ChrArrCheck(wchar[])";
	aStrLp1*    = "int32 [RTS]CP_rts::chrArrLplus1(wchar[])";
	aaStrCmp*   = "int32 [RTS]CP_rts::strCmp(wchar[],wchar[])";
	aaStrCopy*  = "void  [RTS]CP_rts::Stringify(wchar[],wchar[])";
	caseMesg    = 
"class System.String  [RTS]CP_rts::caseMesg(int32)";
	withMesg    = 
"class System.String  [RTS]CP_rts::withMesg(class System.Object)";
	mkExcept    = 
"instance void [mscorlib]System.Exception::.ctor(class System.String)";

   CONST
	chs2Str*    =  "class System.String [RTS]CP_rts::mkStr(wchar[])";
	(* various CPJ-specific concatenation helpers *)
	CPJstrCatAA*   = 
"class System.String [RTS]CP_rts::aaToStr(wchar[],wchar[])";
	CPJstrCatSA*   = 
"class System.String [RTS]CP_rts::saToStr(class System.String,wchar[])";
	CPJstrCatAS*   = 
"class System.String [RTS]CP_rts::asToStr(wchar[], class System.String)";
	CPJstrCatSS*   = 
"class System.String [RTS]CP_rts::ssToStr(class System.String, class System.String)";

(* ============================================================ *)
(* ============================================================ *)

  TYPE ProcInfo*  = POINTER TO RECORD
		      prId- : D.Scope;  (* mth., prc. or mod.	*)
		      exLb  : INTEGER;  (* exception exit label *)
		      rtLc  : INTEGER;  (* return value local # *)
		      (* ---- depth tracking ------ *)
		      dNum  : INTEGER;	(* current stack depth  *)
		      dMax  : INTEGER;	(* maximum stack depth  *)
		      (* ---- temp-var manager ---- *)
		      lNum  : INTEGER;		(* prog vars	*)
		      tLst  : D.TypeSeq;	(* type list	*)
		      fLst  : D.TypeSeq;	(* free list	*)
		      (* ---- end temp manager ---- *)
		    END;

(* ============================================================ *)

  TYPE IlasmFile* = POINTER TO RECORD
                      file* : GPTextFiles.FILE;
		      srcS* : LitValue.CharOpen; (* source file name   *)
		      clsN* : LitValue.CharOpen; (* current class name *)
		      proc* : ProcInfo;
		      nxtLb : INTEGER;
                    END;

(* ============================================================ *)

  VAR   typeGetE  : ARRAY 15 OF INTEGER;
	typePutE  : ARRAY 15 OF INTEGER;
	typeStInd : ARRAY 15 OF INTEGER;
	typeLdInd : ARRAY 15 OF INTEGER;

  VAR	nmArray   : LitValue.CharOpenSeq;
	fmArray   : LitValue.CharOpenSeq;

  VAR   semi, lPar, rPar, lBrk, 	(* ; ( ) {  *)
	rBrk, dotS, lowL, rfMk, 	(* } . _ &  *)
	atSg, slsh, cmma, sQut,		(* @ / , '  *)
	cln2, 				(* "::"     *)
	vFld,				(* "v$"     *)
	mtPr,				(* "()"     *)
	brks,				(* "[]"     *)
	rtsS, 				(* "RTS"    *)
	fpzS, 				(* "0.0"    *)
	clss,				(* "class"  *)
	vals,				(* "value"  *)
	ouMk : LitValue.CharOpen;	(* "[out]"  *)

  VAR	invalid   : LitValue.CharOpen;
	boxedObj  : LitValue.CharOpen;

(* ============================================================ *)
(*			Constructor Method			*)
(* ============================================================ *)

  PROCEDURE newIlasmFile*() : IlasmFile;
    VAR f : IlasmFile;
  BEGIN
    NEW(f);
    RETURN f;
  END newIlasmFile;

(* ==================================================================== *
 *	      A word on naming for the ILASM version.			*
 * ==================================================================== *
 *  Part one: module-level declarations, in Module Mmm.			*
 *		TYPE Ttt = POINTER TO RECORD ... END;			*
 *	has ilasm class name						*
 *		.class <attr> Mmm.Ttt { ... }				*
 *	Similarly the static procedure					*
 *		PROCEDURE Ppp(); END Ppp;				*
 *	has ilasm method name (inside static class Mmm)			*
 *		.method <attr> void Ppp() {...}				*
 *	which is referenced as						*
 *		Ppp(...)		within the static class, &	*
 *		Mmm::Ppp(...)		elswhere inside the module, &	*
 *		[Mmm]Mmm::Ppp(...)	from outside the module.	*
 *	Likewise, methods bound to Ttt will be referenced as		*
 *		Ppp(...)		inside the dynamic class, &	*
 *		Mmm.Ttt::Ppp(...)	elsewhere inside the module, &	*
 *		[Mmm]Mmm.Ttt::Ppp(...)	from outside the module.	*
 *									*
 * ====================================================================	*
 *  Part two: declarations nested inside procedure Outer (say).		*
 *		PROCEDURE Ppp(); END Ppp;				*
 *	will have ilasm name (inside Mmm)				*
 *		.method <attr> void Outer@Ppp() {...}			*
 *	which is referenced as						*
 *		Outer@Ppp(...)						*
 *	Nested type Ttt will have name					*
 *		.struct(?) <attr> Mmm.Outer@Ttt {...}			*
 *	and cannot have type bound procedures, or be exported.		*
 *									*
 * ====================================================================	*
 *  Where are these names stored?  					*
 *  The principle is: every identifier has its class name stored in	*
 *  in d.scopeNm, and its simple name is stored in d.xName.		*
 *  Thus, for names defined in this module:				*
 * ====================================================================	*
 *	The name for BlkId Mmm is stored in desc.xName, as		*
 *		desc.xName   = "Mmm"					*
 *		desc.scopeNm = "Mmm"					*
 *	The names for PrcId Ppp are stored as				*
 *		desc.xName   = "Ppp"					*
 *		desc.scopeNm = "Mmm"					*
 *	or in the nested case...					*
 *		desc.xName   = "Outer@Ppp"				*
 *		desc.scopeNm = "Mmm"					*
 *	The names for (non-nested) MthId Ppp are stored as		*
 *		desc.xName   = "Ppp"					*
 *		desc.scopeNm = "Mmm.Ttt"				*
 *									*
 * For types, the names are stored thuswise.				*
 *	The name for Record descriptor Ttt will be			*
 *		recT.xName   = "Mmm_Ttt"				*
 *		recT.scopeNm = "Mmm_Ttt"				*
 *	or in the nested case ...					*
 *		recT.xName   = "Mmm_Ppp@Ttt"				*
 *		recT.scopeNm = "Mmm_Ppp@Ttt"				*
 *									*
 * ====================================================================	*
 *  Where are these names stored?  For external names:			*
 * ====================================================================	*
 *	The name for BlkId Mmm is stored in desc.xName, as		*
 *		desc.xName   = "Mmm"					*
 *		desc.scopeNm = "[Mmm]Mmm"				*
 *	The names for PrcId Ppp are stored as				*
 *		desc.xName   = "Ppp"					*
 *		desc.scopeNm = "[Mmm]Mmm"				*
 *	The names for (non-nested) MthId Ppp are stored as		*
 *		desc.xName   = "Ppp"					*
 *		desc.scopeNm = "[Mmm]Mmm_Ttt"				*
 *									*
 * For types, the names are stored thuswise.				*
 *	The name for Record descriptor Ttt will be			*
 *		recT.xName   = "Mmm_Ttt"				*
 *		recT.scopeNm = "[Mmm]Mmm_Ttt"				*
 * ====================================================================	*
 * ====================================================================	*)


(* ============================================================ *)
(*			Some static utilities			*)
(* ============================================================ *)

  PROCEDURE^ MkProcName*(proc : Id.Procs);
  PROCEDURE^ MkAliasName*(typ : Ty.Opaque);
  PROCEDURE^ MkEnumName*(typ : Ty.Enum);
  PROCEDURE^ MkTypeName*(typ : D.Type);
  PROCEDURE^ MkRecName*(typ : Ty.Record);
  PROCEDURE^ MkPtrName*(typ : Ty.Pointer);
  PROCEDURE^ MkVarName*(var : Id.VarId);
  PROCEDURE^ MkBlkName*(mod : Id.BlkId);
  PROCEDURE^ MkIdName*(id : D.Idnt);
  PROCEDURE^ (os : IlasmFile)Comment*(IN s : ARRAY OF CHAR),NEW;
  PROCEDURE^ (os : IlasmFile)PushInt*(num : INTEGER),NEW;
  PROCEDURE^ (os : IlasmFile)GetVar*(id : D.Idnt),NEW;
  PROCEDURE^ (os : IlasmFile)GetVarA*(id : D.Idnt),NEW;
  PROCEDURE^ (os : IlasmFile)Code*(code : INTEGER),NEW;
  PROCEDURE^ (os : IlasmFile)CodeS*(code : INTEGER; IN str : ARRAY OF CHAR),NEW;
  PROCEDURE^ (os : IlasmFile)LoadArg*(ord : INTEGER),NEW;
  PROCEDURE^ (os : IlasmFile)LoadLocalA*(ord : INTEGER),NEW;
  PROCEDURE^ (os : IlasmFile)LoadLocal*(ord : INTEGER),NEW;
  PROCEDURE^ (os : IlasmFile)StoreLocal*(ord : INTEGER),NEW;
  PROCEDURE^ (os : IlasmFile)newLabel*() : INTEGER,NEW;
  PROCEDURE^ (os : IlasmFile)Locals*(),NEW;
  PROCEDURE^ (os : IlasmFile)CodeLb*(code, i2 : INTEGER),NEW;
  PROCEDURE^ (os : IlasmFile)MkNewRecord*(typ : Ty.Record),NEW;
  PROCEDURE^ (os : IlasmFile)MkFixedArray*(arTp : Ty.Array),NEW;
  PROCEDURE^ (os : IlasmFile)DefLabC*(l : INTEGER; IN c : ARRAY OF CHAR),NEW;
  PROCEDURE^ (os : IlasmFile)DecTemp(ord : INTEGER),NEW;
  PROCEDURE^ (os : IlasmFile)PutElem*(typ : D.Type),NEW;
  PROCEDURE^ (os : IlasmFile)GetElem*(typ : D.Type),NEW;

(* ------------------------------------------------------------ *)

  PROCEDURE takeAdrs*(i : Id.ParId) : BOOLEAN;
  (*  A parameter needs to have its address taken iff		*)
  (*  * Param Mode is VAL & FALSE				*)
  (*  * Param Mode is VAR & type is value class or scalar	*)
  (*  * Param Mode is OUT & type is value class or scalar	*)
  (*  * Param Mode is IN  & type is value class			*)
  (*    (IN Scalars get treated as VAL on the caller side) 	*)
    VAR type : D.Type;
  BEGIN
    IF i.parMod = D.val THEN RETURN FALSE END;
    type := i.type;
    WITH type : Ty.Array  DO RETURN FALSE;
    |    type : Ty.Record DO RETURN ~(D.clsTp IN type.xAttr);
    ELSE (* scalar type *)   RETURN i.parMod # D.in;
    END;
  END takeAdrs;

(* ------------------------------------------------------------ *)

  PROCEDURE hasValueRep*(type : D.Type) : BOOLEAN;
  BEGIN
    WITH type : Ty.Array  DO	RETURN FALSE;
    |    type : Ty.Record DO	RETURN ~(D.clsTp IN type.xAttr);
    ELSE (* scalar type *)	RETURN TRUE;
    END;
  END hasValueRep;

(* ------------------------------------------------------------ *)

  PROCEDURE isValRecord*(type : D.Type) : BOOLEAN;
  BEGIN
    WITH type : Ty.Array  DO	RETURN FALSE;
    |    type : Ty.Record DO	RETURN ~(D.clsTp IN type.xAttr);
    ELSE (* scalar type *)	RETURN FALSE;
    END;
  END isValRecord;

(* ============================================================ *)
(*			ProcInfo Methods			*)
(* ============================================================ *)

  PROCEDURE newProcInfo*(proc : D.Scope) : ProcInfo;
    VAR p : ProcInfo;
	i : INTEGER;
  BEGIN
   (*
    *  Assert: the locals have already been numbered
    *	       by a call to NumberLocals(), and 
    *	       rtsFram has been set accordingly.
    *)
    NEW(p);
    p.prId := proc;
    WITH proc : Id.Procs DO
      p.lNum := proc.rtsFram;
      IF p.lNum > 0 THEN
	D.InitTypeSeq(p.tLst, p.lNum * 2);	(* the (t)ypeList *)
	D.InitTypeSeq(p.fLst, p.lNum * 2);	(* the (f)reeList *)
	FOR i := 0 TO p.lNum-1 DO 
	  D.AppendType(p.tLst, NIL);
	  D.AppendType(p.fLst, NIL);
	END;
      END;
    ELSE (* Id.BlkId *)
      p.lNum := 0;
    END;
    p.dNum := 0;
    p.dMax := 0;
    p.rtLc := -1;
    RETURN p;
  END newProcInfo;

(* ------------------------------------------------------------ *)

  PROCEDURE (info : ProcInfo)newLocal*(typ : D.Type) : INTEGER,NEW;
    VAR ord : INTEGER;
  BEGIN
   (*
    *   We try to find a previously allocated, but
    *   currently free slot of the identical type.
    *)
    FOR ord := info.lNum TO info.tLst.tide-1 DO
      IF typ.equalType(info.fLst.a[ord]) THEN
	info.fLst.a[ord] := NIL;	(* mark ord as used *)
	RETURN ord;
      END;
    END;
   (* Free slot of correct type not found *)
    ord := info.tLst.tide;
    D.AppendType(info.tLst, typ);
    D.AppendType(info.fLst, NIL);
    RETURN ord;
  END newLocal;

(* ------------------------------------------------------------ *)

  PROCEDURE (info : ProcInfo)ReleaseLocal*(ord : INTEGER),NEW;
  BEGIN
    info.fLst.a[ord] := info.tLst.a[ord];
  END ReleaseLocal;

(* ------------------------------------------------------------ *)

  PROCEDURE (info : ProcInfo)numLocals*() : INTEGER,NEW;
  BEGIN
    RETURN info.tLst.tide;
  END numLocals;

(* ------------------------------------------------------------ *)

  PROCEDURE (info : ProcInfo)SetDepth*(d : INTEGER),NEW;
  BEGIN
    info.dNum := d;
  END SetDepth;

(* ------------------------------------------------------------ *)

  PROCEDURE (info : ProcInfo)getDepth*() : INTEGER,NEW;
  BEGIN
    RETURN info.dNum;
  END getDepth;

(* ------------------------------------------------------------ *)

  PROCEDURE (info : ProcInfo)markTop*() : SET,NEW;
  (*
   *  Returns the set of (first 32) free typed temp-var slots.
   *  The effect of not considering ALL the temporaries is to
   *  miss out on some optimization of storage. Temporary
   *  management is still correct.
   *)
    VAR result : SET;
	index  : INTEGER;
  BEGIN
    result := {};
    FOR index := 0 TO 31 DO
      IF index = info.tLst.tide THEN RETURN result END;
      IF info.fLst.a[index] # NIL THEN INCL(result, index) END;
    END;
    RETURN result;
  END markTop;

(* ------------------------------------------------------------ *)

  PROCEDURE (info : ProcInfo)ReleaseAll*(m : SET),NEW;
    VAR index : INTEGER;
  BEGIN
    FOR index := 0 TO 31 DO
      IF index = info.tLst.tide THEN RETURN END;
      IF index IN m THEN info.fLst.a[index] := info.tLst.a[index] END;
    END;
  END ReleaseAll;

(* ============================================================ *)
(*			Private Methods				*)
(* ============================================================ *)

  PROCEDURE (os : IlasmFile)CatStr(IN str : ARRAY OF CHAR),NEW;
  BEGIN
    GPTextFiles.WriteNChars(os.file, str, LEN(str$));
  END CatStr;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)Tstring(IN str : ARRAY OF CHAR),NEW;
   (* TAB, then string *)
  BEGIN
    GPTextFiles.WriteChar(os.file, "	");
    GPTextFiles.WriteNChars(os.file, str, LEN(str$));
  END Tstring;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)Bstring(IN str : ARRAY OF CHAR),NEW;
   (* BLANK, then string *)
  BEGIN
    GPTextFiles.WriteChar(os.file, " ");
    GPTextFiles.WriteNChars(os.file, str, LEN(str$));
  END Bstring;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)Tint(int : INTEGER),NEW;
   (* TAB, then int *)
  BEGIN
    GPTextFiles.WriteChar(os.file, "	");
    GPText.WriteInt(os.file, int, 1);
  END Tint;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)Tlong(long : LONGINT),NEW;
   (* TAB, then long *)
  BEGIN
    GPTextFiles.WriteChar(os.file, "	");
    GPText.WriteLong(os.file, long, 1);
  END Tlong;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)QuoteStr(IN str : ARRAY OF CHAR),NEW;
    VAR ix : INTEGER;
	ch : CHAR;
  BEGIN
    ix := 0;
    ch := str[0];
    GPTextFiles.WriteChar(os.file, '"');
    WHILE ch # 0X DO
      CASE ch OF
      | "\",'"' : GPTextFiles.WriteChar(os.file, "\");
		  GPTextFiles.WriteChar(os.file, ch);
      | 9X	: GPTextFiles.WriteChar(os.file, "\");
		  GPTextFiles.WriteChar(os.file, "t");
      | 0AX	: GPTextFiles.WriteChar(os.file, "\");
		  GPTextFiles.WriteChar(os.file, "n");
      ELSE
	GPTextFiles.WriteChar(os.file, ch);
      END;
      INC(ix);
      ch := str[ix];
    END;
    GPTextFiles.WriteChar(os.file, '"');
  END QuoteStr;

(* ============================================================ *)

  PROCEDURE cat2(i,j : LitValue.CharOpen) : LitValue.CharOpen;
  BEGIN
    LitValue.ResetCharOpenSeq(nmArray);
    LitValue.AppendCharOpen(nmArray, i);
    LitValue.AppendCharOpen(nmArray, j);
    RETURN LitValue.arrayCat(nmArray); 
  END cat2;

(* ============================================================ *)

  PROCEDURE cat3(i,j,k : LitValue.CharOpen) : LitValue.CharOpen;
  BEGIN
    LitValue.ResetCharOpenSeq(nmArray);
    LitValue.AppendCharOpen(nmArray, i);
    LitValue.AppendCharOpen(nmArray, j);
    LitValue.AppendCharOpen(nmArray, k);
    RETURN LitValue.arrayCat(nmArray); 
  END cat3;

(* ============================================================ *)

  PROCEDURE cat4(i,j,k,l : LitValue.CharOpen) : LitValue.CharOpen;
  BEGIN
    LitValue.ResetCharOpenSeq(nmArray);
    LitValue.AppendCharOpen(nmArray, i);
    LitValue.AppendCharOpen(nmArray, j);
    LitValue.AppendCharOpen(nmArray, k);
    LitValue.AppendCharOpen(nmArray, l);
    RETURN LitValue.arrayCat(nmArray); 
  END cat4;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)Prefix(code : INTEGER),NEW;
  BEGIN
    GPTextFiles.WriteChar(os.file, "	");
    GPTextFiles.WriteNChars(os.file,Asm.op[code],LEN(Asm.op[code]$));
  END Prefix;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)Adjust(delta : INTEGER),NEW;
  BEGIN
    INC(os.proc.dNum, delta); 
    IF os.proc.dNum > os.proc.dMax THEN os.proc.dMax := os.proc.dNum END;
    IF CompState.verbose THEN
      os.CatStr(" //");
      GPText.WriteInt(os.file, os.proc.dNum, 0);
      GPText.WriteInt(os.file, os.proc.dMax, 0);
    END;
    GPTextFiles.WriteEOL(os.file);
  END Adjust;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)Suffix(code : INTEGER),NEW;
  BEGIN
    os.Adjust(Asm.dl[code]);
  END Suffix;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)Access(acc : SET),NEW;
    VAR att : INTEGER;
  BEGIN
    GPTextFiles.WriteChar(os.file, " ");
    FOR att := 0 TO Asm.maxAttIndex DO
      IF att IN acc THEN
	GPText.WriteString(os.file, Asm.access[att]);
	GPTextFiles.WriteChar(os.file, ' ');
      END;
    END;
  END Access;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)RefLab(l : INTEGER),NEW;
  BEGIN
    GPTextFiles.WriteChar(os.file, "	");
    GPTextFiles.WriteChar(os.file, "l");
    GPTextFiles.WriteChar(os.file, "b");
    GPText.WriteInt(os.file, l, 1);
  END RefLab;

(* ------------------------------------	*)

  PROCEDURE (os : IlasmFile)LstLab*(l : INTEGER),NEW;
  BEGIN
    GPTextFiles.WriteChar(os.file, ",");
    GPTextFiles.WriteEOL(os.file);
    os.CatStr("		lb");
    GPText.WriteInt(os.file, l, 1);
  END LstLab;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)SwitchHead*(),NEW;
  BEGIN
    os.Prefix(Asm.opc_switch);
    os.Tstring("( // dispatch table");
  END SwitchHead;

(* ------------------------------------	*)

  PROCEDURE (os : IlasmFile)SwitchTail*(),NEW;
  BEGIN
    GPTextFiles.WriteChar(os.file, ")");
    os.Adjust(-1);
  END SwitchTail;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)Idnt(idD : D.Idnt),NEW;
  BEGIN
    GPTextFiles.WriteChar(os.file, "'");
    GPText.WriteString(os.file, D.getName.ChPtr(idD));
    GPTextFiles.WriteChar(os.file, "'");
  END Idnt;

(* ------------------------------------	*)

  PROCEDURE (os : IlasmFile)SQuote(str : LitValue.CharOpen),NEW;
  BEGIN
    GPTextFiles.WriteChar(os.file, "'");
    GPText.WriteString(os.file, str);
    GPTextFiles.WriteChar(os.file, "'");
  END SQuote;

(* ------------------------------------	*)

  PROCEDURE (os : IlasmFile)TsQuote(str : LitValue.CharOpen),NEW;
  BEGIN
    GPTextFiles.WriteChar(os.file, "	");
    os.SQuote(str);
  END TsQuote;

(* ------------------------------------	*)

  PROCEDURE (os : IlasmFile)Tidnt(idD : D.Idnt),NEW;
  BEGIN
    GPTextFiles.WriteChar(os.file, "	");
    os.Idnt(idD);
  END Tidnt;

(* ------------------------------------	*)

  PROCEDURE (os : IlasmFile)Bidnt(idD : D.Idnt),NEW;
  BEGIN
    GPTextFiles.WriteChar(os.file, " ");
    os.Idnt(idD);
  END Bidnt;

(* ------------------------------------	*)

  PROCEDURE (os : IlasmFile)PIdnt(idD : Id.Procs),NEW;
    VAR clsNam : LitValue.CharOpen;
  BEGIN
    IF idD.scopeNm = NIL THEN MkProcName(idD) END;
    GPTextFiles.WriteChar(os.file, " ");
    GPText.WriteString(os.file, idD.scopeNm);
    WITH idD : Id.PrcId DO clsNam := idD.clsNm;
    |    idD : Id.MthId DO clsNam := idD.bndType.xName;
    END;
    GPText.WriteString(os.file, clsNam);
    GPText.WriteString(os.file, cln2);
    os.SQuote(idD.prcNm);
  END PIdnt;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)TypeTag(typ : D.Type),NEW;
  BEGIN
    IF typ.xName = NIL THEN MkTypeName(typ) END;
    WITH typ : Ty.Base DO
	GPText.WriteString(os.file, typ.xName);
    | typ : Ty.Array DO
	os.TypeTag(typ.elemTp);
	GPText.WriteString(os.file, brks);
    | typ : Ty.Record DO
	IF ~(D.clsTp IN typ.xAttr) THEN os.CatStr(vals) END;
	GPText.WriteString(os.file, clss);
	GPText.WriteString(os.file, typ.scopeNm);
    | typ : Ty.Pointer DO
	IF isValRecord(typ.boundTp) THEN
	 (* 
	  *  This is a pointer to a value class, which has a
	  *  runtime representation as a boxed-class reference.
	  *)
	  GPText.WriteString(os.file, clss);
	  GPText.WriteString(os.file, typ.xName);
	ELSE
	  os.TypeTag(typ.boundTp);
	END;
    | typ : Ty.Opaque DO
	GPText.WriteString(os.file, clss);
	GPText.WriteString(os.file, typ.xName);
    | typ : Ty.Enum DO
	GPText.WriteString(os.file, vals);
	GPText.WriteString(os.file, clss);
	GPText.WriteString(os.file, typ.xName);
    END;
  END TypeTag;

(* ------------------------------------	*)

  PROCEDURE typeName(typ : D.Type) : LitValue.CharOpen;
  BEGIN
    IF typ.xName = NIL THEN MkTypeName(typ) END;
    WITH typ : Ty.Base DO
	RETURN typ.xName;
    | typ : Ty.Array DO
	RETURN typ.xName;
    | typ : Ty.Record DO
	RETURN typ.scopeNm;
    | typ : Ty.Pointer DO
	RETURN typ.xName;
    | typ : Ty.Opaque DO
	RETURN typ.xName;
    | typ : Ty.Enum DO
	RETURN typ.xName;
    END;
  END typeName;

(* ------------------------------------	*)

  PROCEDURE boxedName*(typ : Ty.Record) : LitValue.CharOpen;
  BEGIN
    IF typ.xName = NIL THEN MkRecName(typ) END;
    RETURN cat3(typ.idnt.dfScp.scopeNm, boxedObj, typ.xName);
  END boxedName;

(* ------------------------------------	*)

  PROCEDURE MkTypeName(typ : D.Type);
  BEGIN
    ASSERT(typ.xName = NIL);
    WITH typ : Ty.Array   DO typ.xName := cat2(typeName(typ.elemTp), brks);
      |  typ : Ty.Record  DO MkRecName(typ);
      |  typ : Ty.Pointer DO MkPtrName(typ);
      |  typ : Ty.Opaque  DO MkAliasName(typ);
      |  typ : Ty.Enum    DO MkEnumName(typ);
    END;
  END MkTypeName;

(* ------------------------------------	*)

  PROCEDURE (os : IlasmFile)TtypeTag(typ : D.Type),NEW;
  BEGIN
    GPTextFiles.WriteChar(os.file, "	");
    os.TypeTag(typ);
  END TtypeTag;

(* ------------------------------------	*)

  PROCEDURE (os : IlasmFile)TtypeNam(typ : D.Type),NEW;
  BEGIN
    GPTextFiles.WriteChar(os.file, "	");
    GPText.WriteString(os.file, typeName(typ));
  END TtypeNam;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)RetType(typ : Ty.Procedure; pId : Id.Procs),NEW;
  BEGIN
    IF typ.retType = NIL THEN
      os.CatStr("void");
    ELSIF (pId IS Id.MthId) & (Id.covar IN pId(Id.MthId).mthAtt) THEN
     (*
      *  This is a method with a covariant return type. We must
      *  erase the declared type, substituting the non-covariant
      *  upper-bound. Calls will cast the result to the real type.
      *)
      os.TypeTag(pId.retTypBound());
    ELSE
      os.TypeTag(typ.retType);
    END;
  END RetType;

(* ============================================================ *)
(*			Exported Methods			*)
(* ============================================================ *)

  PROCEDURE (os : IlasmFile)newLabel*() : INTEGER,NEW;
  BEGIN
    INC(os.nxtLb); RETURN os.nxtLb;
  END newLabel;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)getLabelRange*(num : INTEGER) : INTEGER,NEW;
    VAR start : INTEGER;
  BEGIN
    start := os.nxtLb + 1;
    INC(os.nxtLb, num); 
    RETURN start;
  END getLabelRange;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)Blank*(),NEW;
  BEGIN
    GPTextFiles.WriteEOL(os.file);
  END Blank;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)Separator(c : CHAR; i : INTEGER),NEW;
  BEGIN
    GPTextFiles.WriteChar(os.file, c);
    GPTextFiles.WriteEOL(os.file);
    WHILE i > 0 DO GPTextFiles.WriteChar(os.file, "	"); DEC(i) END;
  END Separator;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)OpenBrace*(i : INTEGER),NEW;
  BEGIN
    WHILE i > 0 DO GPTextFiles.WriteChar(os.file, " "); DEC(i) END;
    GPTextFiles.WriteChar(os.file, "{");
    GPTextFiles.WriteEOL(os.file);
  END OpenBrace;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)CloseBrace*(i : INTEGER),NEW;
  BEGIN
    WHILE i > 0 DO GPTextFiles.WriteChar(os.file, " "); DEC(i) END;
    GPTextFiles.WriteChar(os.file, "}");
    GPTextFiles.WriteEOL(os.file);
  END CloseBrace;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)Directive*(dir : INTEGER),NEW;
  BEGIN
    os.CatStr(Asm.dirStr[dir]);
    GPTextFiles.WriteEOL(os.file);
  END Directive;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)DirectiveS*(dir : INTEGER;
				   IN str : ARRAY OF CHAR),NEW;
  BEGIN
    os.CatStr(Asm.dirStr[dir]);
    os.Bstring(str);
    GPTextFiles.WriteEOL(os.file);
  END DirectiveS;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)DirectiveIS*(dir : INTEGER;
				        att : SET;
				     IN str : ARRAY OF CHAR),NEW;
  BEGIN
    os.CatStr(Asm.dirStr[dir]);
    os.Access(att);
    GPTextFiles.WriteNChars(os.file, str, LEN(str$));
    GPTextFiles.WriteEOL(os.file);
  END DirectiveIS;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)DirectiveISS*(dir : INTEGER;
				         att : SET;
				      IN s1  : ARRAY OF CHAR;
				      IN s2  : ARRAY OF CHAR),NEW;
  BEGIN
    os.CatStr(Asm.dirStr[dir]);
    os.Access(att);
    GPTextFiles.WriteNChars(os.file, s1, LEN(s1$));
    GPTextFiles.WriteNChars(os.file, s2, LEN(s2$));
    IF dir = Asm.dot_method THEN os.Bstring(managedStr) END;
    GPTextFiles.WriteEOL(os.file);
  END DirectiveISS;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)MethodDecl*(attr : SET; proc : Id.Procs),NEW;
    CONST inlineMax = 2;
    VAR indx : INTEGER;
	prcT : Ty.Procedure;
	parD : Id.ParId;
	long : BOOLEAN;
  BEGIN
    prcT := proc.type(Ty.Procedure);
    os.CatStr(Asm.dirStr[Asm.dot_method]);
    os.Access(attr);
    os.RetType(prcT, proc);
    GPTextFiles.WriteChar(os.file, " ");
    GPTextFiles.WriteChar(os.file, "'");
    os.CatStr(proc.prcNm);
    GPTextFiles.WriteChar(os.file, "'");
    GPTextFiles.WriteChar(os.file, "(");
    long := prcT.formals.tide > inlineMax;
    IF long THEN os.Separator(' ', 2) END;
    FOR indx := 0 TO prcT.formals.tide-1 DO
      parD := prcT.formals.a[indx];
      IF long THEN
        IF indx > 0 THEN os.Separator(',', 2) END;
        IF parD.boxOrd = D.out THEN os.CatStr(ouMk) END;
        os.TypeTag(parD.type);
        IF parD.boxOrd # D.val THEN os.CatStr(rfMk) END;
        os.Tidnt(parD); 
      ELSE
        IF indx > 0 THEN os.CatStr(cmma) END;
        IF parD.boxOrd = D.out THEN os.CatStr(ouMk) END;
        os.TypeTag(parD.type);
        IF parD.boxOrd # D.val THEN os.CatStr(rfMk) END;
        os.Bidnt(parD);
      END;
    END;
    GPTextFiles.WriteChar(os.file, ")");
    os.Bstring(managedStr);
    GPTextFiles.WriteEOL(os.file);
  END MethodDecl;

(* --------------------------------------------	*)

  PROCEDURE (os : IlasmFile)ExternList*(),NEW;
    VAR idx : INTEGER;
	blk : Id.BlkId;
   (* -----------------------------------------	*)
    PROCEDURE Assembly(fl : IlasmFile; bk : Id.BlkId);
      VAR ix : INTEGER;
	  ch : CHAR;
    BEGIN
      IF (D.isCs IN bk.xAttr) OR
	 (D.isJv IN bk.xAttr) THEN
	IF bk.scopeNm[0] # '[' THEN 
				RTS.Throw("bad extern name "+bk.scopeNm^) END;
	ix := 1;
	ch := bk.scopeNm[ix];
	WHILE (ch # 0X) & (ch # ']') DO
	  GPTextFiles.WriteChar(fl.file, ch);
	  INC(ix);
	  ch := bk.scopeNm[ix];
	END;
      ELSE
	fl.CatStr(bk.xName);
      END;
    END Assembly;
   (* -----------------------------------------	*)
  BEGIN
   (*
    *   It is empirically established that all the
    *   .assembly extern declarations must come at
    *   the beginning of the ILASM file, at once.
    *)
    FOR idx := 0 TO CompState.impSeq.tide-1 DO
      blk := CompState.impSeq.a[idx](Id.BlkId);
(*
IF ~(D.rtsMd IN blk.xAttr) & ~(D.need IN blk.xAttr) THEN
Console.WriteString("Ignoring ");
MkBlkName(blk);
D.DoXName(0, blk.xName);
END;
 *)
      IF ~(D.rtsMd IN blk.xAttr) & (D.need IN blk.xAttr) THEN
	MkBlkName(blk);
	os.CatStr(Asm.dirStr[Asm.dot_assembly]);
	os.Access(Asm.att_extern);
	Assembly(os, blk);
	os.CatStr(" {}");
	GPTextFiles.WriteEOL(os.file);
      END;
    END;
  END ExternList;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)Comment*(IN s : ARRAY OF CHAR),NEW;
  BEGIN
    os.CatStr("// ");
    GPTextFiles.WriteNChars(os.file, s, LEN(s$));
    GPTextFiles.WriteEOL(os.file);
  END Comment;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)CommentT*(IN s : ARRAY OF CHAR),NEW;
  BEGIN
    os.CatStr("//	");
    GPTextFiles.WriteNChars(os.file, s, LEN(s$));
    GPTextFiles.WriteEOL(os.file);
  END CommentT;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)DefLab*(l : INTEGER),NEW;
  BEGIN
    GPTextFiles.WriteChar(os.file, "l");
    GPTextFiles.WriteChar(os.file, "b");
    GPText.WriteInt(os.file, l, 1);
    GPTextFiles.WriteChar(os.file, ":");
    GPTextFiles.WriteEOL(os.file);
  END DefLab;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)DefLabC*(l : INTEGER; IN c : ARRAY OF CHAR),NEW;
  BEGIN
    GPTextFiles.WriteChar(os.file, "l");
    GPTextFiles.WriteChar(os.file, "b");
    GPText.WriteInt(os.file, l, 1);
    GPTextFiles.WriteChar(os.file, ":");
    GPTextFiles.WriteChar(os.file, "	");
    os.Comment(c);
  END DefLabC;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)Code*(code : INTEGER),NEW;
  BEGIN
    os.Prefix(code);
    os.Suffix(code);
  END Code;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)CodeI*(code,int : INTEGER),NEW;
  BEGIN
    os.Prefix(code);
    os.Tint(int);
    os.Suffix(code);
  END CodeI;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)CodeT*(code : INTEGER; type : D.Type),NEW;
  BEGIN
    os.Prefix(code);
    os.TtypeTag(type);
    os.Suffix(code);
  END CodeT;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)CodeTn*(code : INTEGER; type : D.Type),NEW;
  BEGIN
    os.Prefix(code);
    os.TtypeNam(type);
    os.Suffix(code);
  END CodeTn;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)CodeL*(code : INTEGER; long : LONGINT),NEW;
  BEGIN
    os.Prefix(code);
    os.Tlong(long);
    os.Suffix(code);
  END CodeL;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)CodeR*(code : INTEGER; real : REAL),NEW;
    VAR nam : ARRAY 64 OF CHAR;
  BEGIN
    os.Prefix(code);
    RTS.RealToStr(real, nam);
    os.Tstring(nam$);
    os.Suffix(code);
  END CodeR;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)CodeLb*(code,i2 : INTEGER),NEW;
  BEGIN
    os.Prefix(code);
    os.RefLab(i2);
    os.Suffix(code);
  END CodeLb;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)CodeII*(code,i1,i2 : INTEGER),NEW;
  BEGIN
    os.Prefix(code);
    os.Tint(i1);
    os.Tint(i2);
    os.Suffix(code);
  END CodeII;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)CodeS*(code : INTEGER; IN str : ARRAY OF CHAR),NEW;
  BEGIN
    os.Prefix(code);
    os.Tstring(str);
    os.Suffix(code);
  END CodeS;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)StaticCall*(IN s : ARRAY OF CHAR; d : INTEGER),NEW;
  BEGIN
    os.Prefix(Asm.opc_call);
    os.Tstring(s);
    os.Adjust(d);
  END StaticCall;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)Try*(),NEW;
    VAR retT : D.Type;
  BEGIN
    retT := os.proc.prId.type.returnType();
    os.Directive(Asm.dot_try);
    os.OpenBrace(4);
    os.proc.exLb := os.newLabel();
    IF retT # NIL THEN os.proc.rtLc := os.proc.newLocal(retT) END;
  END Try;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)Catch*(proc : Id.Procs),NEW;
  BEGIN
    os.CloseBrace(4);
    os.CatStr(catchStr);
    GPTextFiles.WriteEOL(os.file);
    os.OpenBrace(4);
    os.StoreLocal(proc.except.varOrd);
  END Catch;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)EndCatch*(),NEW;
  BEGIN
    os.CloseBrace(4);
    os.DefLab(os.proc.exLb);
    IF os.proc.rtLc # -1 THEN os.LoadLocal(os.proc.rtLc) END;
    os.Code(Asm.opc_ret);
  END EndCatch;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)DoReturn*(),NEW;
  BEGIN
    IF os.proc.exLb = 0 THEN
      os.Code(Asm.opc_ret);
      IF os.proc.prId.type.returnType() # NIL THEN DEC(os.proc.dNum) END;
    ELSE
      IF os.proc.rtLc # -1 THEN os.StoreLocal(os.proc.rtLc) END;
      os.CodeLb(Asm.opc_leave, os.proc.exLb);
    END;
  END DoReturn;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)CopyCall*(typ : Ty.Record),NEW;
  BEGIN
    os.Prefix(Asm.opc_call);
    os.Tstring(initPrefix);
    os.Bstring(typ.scopeNm);
    os.CatStr("::__copy__(");
    os.TypeTag(typ);
    GPTextFiles.WriteChar(os.file, ")");
    os.Adjust(-2);
  END CopyCall;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)PushStr*(IN str : ARRAY OF CHAR),NEW;
  (* Use target quoting conventions for the literal string *)
  BEGIN
    os.Prefix(Asm.opc_ldstr);
    GPTextFiles.WriteChar(os.file, "	");
    os.QuoteStr(str);
    os.Adjust(1);
  END PushStr;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)CallIT*(code : INTEGER; 
				    proc : Id.Procs; 
				    type : Ty.Procedure),NEW;
  BEGIN
    os.Prefix(code);
    GPTextFiles.WriteChar(os.file, "	");
   (*
    *   For static calls to procedures we want
    *	    call  <ret-type> <idnt>(<signature>)
    *   for static calls to final or super methods, we need
    *	    call  instance <ret-type> <idnt>(<signature>)
    *   for calls to type-bound methods that are not final
    *	    callvirt  instance <ret-type> <idnt>(<signature>)
    *)
    IF proc IS Id.MthId THEN os.CatStr("instance ") END;
    os.RetType(type, proc);
    os.PIdnt(proc);
    os.CatStr(type.xName);
    os.Adjust(type.retN - type.argN);	
  END CallIT;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)CallCT*(proc : Id.Procs; 
				    type : Ty.Procedure),NEW;
  BEGIN
    os.Prefix(Asm.opc_newobj);
    os.Tstring(initPrefix);
    os.PIdnt(proc);
    os.CatStr(type.xName);
    os.Adjust(type.retN - type.argN);	
  END CallCT;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)MkFixedArray*(arTp : Ty.Array),NEW;
    VAR cTmp : INTEGER;			(* card'ty of this dim. *)
	aTmp : INTEGER;			(* array reference temp *)
	labl : INTEGER;
	elTp : D.Type;
	aLen : INTEGER;
  BEGIN
    ASSERT(arTp.length # 0);
    elTp := arTp.elemTp;
    aLen := arTp.length;
    os.PushInt(aLen);
    os.CodeTn(Asm.opc_newarr, elTp);
   (*
    *   Do we need an initialization loop?
    *)
    IF ~hasValueRep(elTp) THEN
      labl := os.newLabel();
      cTmp := os.proc.newLocal(G.intTp);
      aTmp := os.proc.newLocal(arTp);
      os.StoreLocal(aTmp);		(* (top)...		*)
      os.PushInt(aLen);
      os.StoreLocal(cTmp);
     (*
      *  Now the allocation loop
      *)
      os.DefLab(labl);
      os.DecTemp(cTmp);
      os.LoadLocal(aTmp);
      os.LoadLocal(cTmp);
      WITH elTp : Ty.Array DO
	  os.MkFixedArray(elTp);
      | elTp : Ty.Record DO
	  os.MkNewRecord(elTp);
      END;				(* (top)elem,ix,ref,...	*)
      os.PutElem(elTp);
     (*
      *  Now the termination test
      *)
      os.LoadLocal(cTmp);
      os.CodeLb(Asm.opc_brtrue, labl);
      os.LoadLocal(aTmp);
      os.proc.ReleaseLocal(cTmp);
      os.proc.ReleaseLocal(aTmp);
    END;
  END MkFixedArray;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)MkOpenArray*(arTp : Ty.Array),NEW;
    VAR lens : ARRAY 32 OF INTEGER;
	elTp : D.Type;
   (* -----------------------------------------	*)
    PROCEDURE GetLengths(os  : IlasmFile;
			 dim : INTEGER;
			 typ : D.Type;
		     VAR lAr : ARRAY OF INTEGER);
      VAR tmp : INTEGER;
    BEGIN
      ASSERT(dim < 31);
      WITH typ : Ty.Array DO
	IF typ.length = 0 THEN	(* another open dimension *)
          tmp := os.proc.newLocal(G.intTp);
          lAr[dim] := tmp;
	  os.StoreLocal(tmp);
	  GetLengths(os, dim+1, typ.elemTp, lAr);
	END;
      ELSE
      END;
    END GetLengths;
   (* -----------------------------------------	*)
    PROCEDURE InitLoop(os  : IlasmFile;
		       dim : INTEGER;
		       typ : Ty.Array;
		    IN lAr : ARRAY OF INTEGER);
      VAR aEl : INTEGER;
	  lab : INTEGER;
	  elT : D.Type;
    BEGIN
     (*
      *  Pre  : the uninitialized array is on the stack
      *)
      elT := typ.elemTp;
      IF ~hasValueRep(elT) THEN
	aEl := os.proc.newLocal(typ);
	os.StoreLocal(aEl);
        lab := os.newLabel();
       (*
	*  Start of initialization loop
	*)
	os.DefLab(lab);
       (*
	*  Decrement the loop counter
	*)
	os.DecTemp(lAr[dim]);
       (*
	*  Assign the array element
	*)
	os.LoadLocal(aEl);
	os.LoadLocal(lAr[dim]);
	WITH elT : Ty.Record DO
	    os.MkNewRecord(elT);
	| elT : Ty.Array DO
	    IF elT.length > 0 THEN
	      os.MkFixedArray(elT);
	    ELSE
	      os.LoadLocal(lAr[dim+1]);
	      os.CodeTn(Asm.opc_newarr, elT.elemTp);
	      InitLoop(os, dim+1, elT, lAr);
	    END;
        END;
	os.PutElem(elT);
       (*
	*  Test and branch to loop header
	*)
	os.LoadLocal(lAr[dim]);
	os.CodeLb(Asm.opc_brtrue, lab);
       (*
	*  Reload the original array
	*)
	os.LoadLocal(aEl);
	os.proc.ReleaseLocal(aEl);
	os.proc.ReleaseLocal(lAr[dim]);
      END;
     (*
      *  Post : the initialized array is on the stack
      *)
    END InitLoop;
   (* -----------------------------------------	*)
  BEGIN
    elTp := arTp.elemTp;
    IF (elTp IS Ty.Array) OR (elTp IS Ty.Record) THEN
      GetLengths(os, 0, arTp, lens);
      os.LoadLocal(lens[0]);
      os.CodeTn(Asm.opc_newarr, elTp);
      InitLoop(os, 0, arTp, lens);
    ELSE
      os.CodeTn(Asm.opc_newarr, elTp);
    END; 
  END MkOpenArray;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)PutGetS*(code : INTEGER;
				     blk  : Id.BlkId;
				     fld  : Id.VarId),NEW;
    VAR size : INTEGER;
  (* Emit putstatic and getstatic for static field *)
  BEGIN
    os.Prefix(code);
    os.TtypeTag(fld.type);
    GPTextFiles.WriteChar(os.file, "	");
    IF blk.xName = NIL THEN MkBlkName(blk) END;
    IF fld.varNm = NIL THEN MkVarName(fld) END;
    os.CatStr(blk.scopeNm);
    os.CatStr(fld.clsNm);
    os.CatStr(cln2);
    os.SQuote(fld.varNm);
    os.Suffix(code);
  END PutGetS;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)PutGetFld(code : INTEGER;
				      fTyp : D.Type;
				      rTyp : D.Type;
				      name : LitValue.CharOpen),NEW;
  BEGIN
    os.Prefix(code);
    os.TtypeTag(fTyp);
    os.TtypeNam(rTyp);
    os.CatStr(cln2);
    os.SQuote(name);
    os.Suffix(code);
  END PutGetFld;

(* -------------------------------------------- *)

  PROCEDURE (os : IlasmFile)PutGetF*(code : INTEGER;
				     fld  : Id.FldId),NEW;
    VAR recT : Ty.Record;
  (* Emit putfield and getfield for record field *)
  BEGIN
    recT := fld.recTyp(Ty.Record);
    os.Prefix(code);
    os.TtypeTag(fld.type);
    GPTextFiles.WriteChar(os.file, "	");
    IF fld.fldNm = NIL THEN fld.fldNm := D.getName.ChPtr(fld) END;
   (*
    *  Note the difference here. JVM needs the
    *  static type of the variable, VOS wants
    *  the name of the record with the field.
    *)
    os.CatStr(recT.scopeNm);
    os.CatStr(cln2);
    os.SQuote(fld.fldNm);
    os.Suffix(code);
  END PutGetF;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)MkNewRecord*(typ : Ty.Record),NEW;
    VAR name : LitValue.CharOpen;
  BEGIN
   (*
    *  We need "newobj instance void <name>::.ctor()"
    *)
    IF typ.xName = NIL THEN MkRecName(typ) END;
    IF D.clsTp IN typ.xAttr THEN
      name := typ.scopeNm;
    ELSE
      name := boxedName(typ);
    END;
    os.Prefix(Asm.opc_newobj);
    os.Tstring(initPrefix);
    os.Bstring(name);
    os.CatStr(cln2);
    os.CatStr(initSuffix);
    os.Adjust(1);
  END MkNewRecord;

(* ============================================================ *)
(* ============================================================ *)

  PROCEDURE (os : IlasmFile)InitHead*(typ : Ty.Record),NEW;
  BEGIN
    os.CatStr(Asm.dirStr[Asm.dot_method]);
    os.Tstring(specialStr);
    os.Tstring(initPrefix);
    os.Bstring(initSuffix);
    os.CatStr(managedStr);
    GPTextFiles.WriteEOL(os.file);
    os.CatStr("    {");
    GPTextFiles.WriteEOL(os.file);
   (*
    *   Now we initialize the supertype;
    *)
    os.CommentT("Call supertype constructor");
    os.Code(Asm.opc_ldarg_0);
    os.Prefix(Asm.opc_call);
    IF  (typ # NIL) & 
	(typ.baseTp # NIL) & 
	(typ.baseTp # G.anyRec) THEN
      os.Tstring(initPrefix);
      os.Bstring(typ.baseTp(Ty.Record).scopeNm);
      os.CatStr(cln2);
      os.CatStr(initSuffix);
    ELSE
      os.Tstring(objectInit);
    END;
    os.Adjust(-1);
  END InitHead;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)InitTail*(typ : Ty.Record),NEW;
  BEGIN
    os.Locals();
    os.CatStr("    } // end of method '");
    IF typ # NIL THEN 
      os.CatStr(typ.xName);
    END;
    os.CatStr("::.ctor()'");
    GPTextFiles.WriteEOL(os.file);
  END InitTail;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)CopyHead*(typ : Ty.Record),NEW;
  BEGIN
    os.proc := newProcInfo(NIL);
    os.Comment("standard record copy method");
    os.CatStr(Asm.dirStr[Asm.dot_method]);
    os.Tstring(copyHead);
(* FIX FOR BOXED CLASS COPY *)
    IF ~(D.clsTp IN typ.xAttr) THEN os.CatStr(vals) END;
    os.CatStr(clss);
    os.CatStr(typ.scopeNm);
    IF ~(D.clsTp IN typ.xAttr) THEN os.CatStr(rfMk) END;
    os.CatStr(") ");
    os.CatStr(managedStr);
    GPTextFiles.WriteEOL(os.file);
    os.CatStr("    {");
    GPTextFiles.WriteEOL(os.file);
  END CopyHead;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)CopyTail*(),NEW;
  BEGIN
    os.Locals();
    os.CatStr("    } // end of __copy__ method '");
    GPTextFiles.WriteEOL(os.file);
  END CopyTail;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)MainHead*(),NEW;
  BEGIN
    (* os.proc := newProcInfo(NIL); *)
    os.Comment("Main entry point");
    os.CatStr(Asm.dirStr[Asm.dot_method]);
    os.Tstring(mainStr);
    os.CatStr(managedStr);
    GPTextFiles.WriteEOL(os.file);
    os.OpenBrace(4);
    os.Directive(Asm.dot_entrypoint);
   (*
    *  Save the command-line arguments to the RTS.
    *)
    os.Code(Asm.opc_ldarg_0);
    os.CodeS(Asm.opc_stsfld, putArgStr);
  END MainHead;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)EmitBoxClass*(rec : Ty.Record;
					  att : SET;
					  blk : Id.BlkId),NEW;
    VAR name : LitValue.CharOpen;
	bNam : LitValue.CharOpen;
  BEGIN
    GPTextFiles.WriteEOL(os.file);
    name := cat2(boxedObj, os.clsN);
    bNam := boxedName(rec);
    os.DirectiveIS(Asm.dot_class, att, name);
    os.OpenBrace(2);
    os.CatStr(Asm.dirStr[Asm.dot_field]);
    os.Access(Asm.att_public);
    os.TtypeTag(rec);
    os.Bstring(vFld);
    GPTextFiles.WriteEOL(os.file);
   (*
    *   Emit the no-arg constructor
    *)
    os.Blank();
    os.proc := newProcInfo(blk);
    os.InitHead(rec);
    os.Code(Asm.opc_ret);
    os.InitTail(rec);
   (*
    *   Copies of value classes are always done inline.
    *)
    os.CloseBrace(2);
  END EmitBoxClass;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)MainTail*(),NEW;
  BEGIN
    os.Locals();
    os.CatStr("    } // end of method Main");
    GPTextFiles.WriteEOL(os.file);
  END MainTail;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)ClinitHead*(),NEW;
  BEGIN
    os.CatStr(Asm.dirStr[Asm.dot_method]);
    os.Tstring(specialStr);
    os.CatStr(cctorStr);
    os.CatStr(managedStr);
    GPTextFiles.WriteEOL(os.file);
    os.CatStr("    {");
    GPTextFiles.WriteEOL(os.file);
  END ClinitHead;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)ClinitTail*(),NEW;
  BEGIN
    os.Locals();
    os.CatStr("    } // end of .cctor method");
    GPTextFiles.WriteEOL(os.file);
  END ClinitTail;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)MethodTail*(id : Id.Procs),NEW;
  BEGIN
    os.Locals();
    os.CatStr("    } // end of method ");
    os.CatStr(id.prcNm);
    GPTextFiles.WriteEOL(os.file);
  END MethodTail;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)MkArrayCopy*(arrT : Ty.Array),NEW;
    VAR dims : INTEGER;
        elTp : D.Type;
 (* -----------------------------------	*)
    PROCEDURE PushLengths(os : IlasmFile; aT : Ty.Array);
    BEGIN
      IF aT.elemTp IS Ty.Array THEN
	os.Code(Asm.opc_dup);
	os.Code(Asm.opc_ldc_i4_0);
	os.GetElem(aT.elemTp);
	PushLengths(os, aT.elemTp(Ty.Array));
      END;
      os.Code(Asm.opc_ldlen);
    END PushLengths;
 (* -----------------------------------	*)
  BEGIN
   (*
    *	Assert: we must find the lengths from the runtime 
    *   descriptors. The array to copy is on the top of 
    *   stack, which reads - 	(top) aRef, ...
    *)
    PushLengths(os, arrT);
    os.MkOpenArray(arrT);
  END MkArrayCopy;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)StructInit*(var : D.Idnt),NEW;
    VAR typ : D.Type;
	fld : D.Idnt;
	idx : INTEGER;
   (* -------------------------------------------------	*)
    PROCEDURE Assign(os : IlasmFile; id : D.Idnt);
      VAR md : Id.BlkId;
    BEGIN
      WITH id : Id.LocId DO
	  os.StoreLocal(id.varOrd);
      | id : Id.FldId DO
          os.PutGetF(Asm.opc_stfld, id);
      | id : Id.VarId DO
	  md := id.dfScp(Id.BlkId);
	  os.PutGetS(Asm.opc_stsfld, md, id);
      END;
    END Assign;
   (* -------------------------------------------------	*)
  BEGIN
    os.Comment("initialize " + D.getName.ChPtr(var)^);
   (*
    *  Precondition: var is of a type that needs initialization,
    *)
    typ := var.type;
    WITH typ : Ty.Array DO
	os.MkFixedArray(typ);
	Assign(os, var);
    | typ : Ty.Record DO
	IF D.clsTp IN typ.xAttr THEN
	 (*
	  *  Reference record type
	  *)
	  os.MkNewRecord(typ);
	  Assign(os, var);
	ELSE
	 (*
	  *  Value record type
	  *)
	  os.GetVarA(var);
	  os.CodeTn(Asm.opc_initobj, typ);
       (*
	*    In this version, value types cannot
	*    have embedded value fields which have
	*    a reference representation in .NET
	*(*
	* *  Initialize any embedded fields
	* *)
	* FOR idx := 0 TO typ.fields.tide-1 DO
	*   fld := typ.fields.a[idx];
	*   IF ~hasValueRep(fld.type) THEN
	*     os.GetVarA(var);
	*     os.StructInit(fld);
	*   END;
	* END;
	*)
	END;
    ELSE
      os.Code(Asm.opc_ldnull);
      Assign(os, var);
    END;
  END StructInit;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)PushZero(typ : D.Type),NEW;
    VAR cde : INTEGER;
  BEGIN
    WITH typ : Ty.Base DO
      CASE typ.tpOrd OF
      | Ty.sReaN : os.CodeS(Asm.opc_ldc_r4, fpzS);
      | Ty.realN : os.CodeS(Asm.opc_ldc_r8, fpzS);
      | Ty.lIntN : os.CodeL(Asm.opc_ldc_i8, 0);
      | Ty.charN,
        Ty.sChrN : os.Code(Asm.opc_ldc_i4_0);
      ELSE	   os.Code(Asm.opc_ldc_i4_0);
      END;
    ELSE
      os.Code(Asm.opc_ldnull);
    END;
  END PushZero;

 (* -----------------------------------	*)

  PROCEDURE (os : IlasmFile)ScalarInit(var : D.Idnt),NEW;
    VAR typ : D.Type;
	cde : INTEGER;
  BEGIN
    os.Comment("initialize " + D.getName.ChPtr(var)^);
    typ := var.type;
   (*
    *  Precondition: var is of a scalar type that is referenced
    *)
    os.PushZero(typ);
  END ScalarInit;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)Throw*(),NEW;
  BEGIN
    os.CodeS(Asm.opc_newobj, mkExcept);
    os.Code(Asm.opc_throw);
  END Throw;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)Trap*(IN str : ARRAY OF CHAR),NEW;
  BEGIN
(* Do we need the quotes? *)
    os.PushStr('"' + str + '"');
    os.Throw();
  END Trap;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)CaseTrap*(i : INTEGER),NEW;
  BEGIN
    os.Comment("CaseTrap");
    os.LoadLocal(i);
    os.CodeS(Asm.opc_call, caseMesg);
    os.CodeS(Asm.opc_newobj, mkExcept);
    os.Code(Asm.opc_throw);
  END CaseTrap;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)WithTrap*(id : D.Idnt),NEW;
  BEGIN
    os.Comment("WithTrap " + D.getName.ChPtr(id)^);
    os.GetVar(id);
    os.CodeS(Asm.opc_call, withMesg);
    os.CodeS(Asm.opc_newobj, mkExcept);
    os.Code(Asm.opc_throw);
  END WithTrap;
(* ============================================================ *)

  PROCEDURE (os : IlasmFile)Header*(IN str : ARRAY OF CHAR),NEW;
    VAR date : ARRAY 64 OF CHAR;
  BEGIN
    os.srcS := LitValue.strToCharOpen("'" + str + "'");
    RTS.GetDateString(date);
    os.Comment("ILASM output produced by GPCP compiler (" +
					RTS.defaultTarget + " version)");
    os.Comment("at date: " + date);
    os.Comment("from source file <" + str + '>');
  END Header;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)EmitField*(id : D.Idnt; att : SET),NEW;
    VAR nm : LitValue.CharOpen;
  BEGIN
    IF id.vMod # D.prvMode THEN
      att := att + Asm.att_public;
    ELSE
      att := att + Asm.att_assembly;
    END;
    MkIdName(id);
    os.CatStr(Asm.dirStr[Asm.dot_field]);
    os.Access(att);
    os.TtypeTag(id.type);
    WITH id : Id.FldId DO nm := id.fldNm;
    | id : Id.VarId DO    nm := id.varNm;
    END;
    os.TsQuote(nm);
    GPTextFiles.WriteEOL(os.file);
  END EmitField;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)Line*(nm : INTEGER),NEW;
  BEGIN
    os.CatStr(Asm.dirStr[Asm.dot_line]); 
    os.Tint(nm);
    os.Tstring(os.srcS);
    GPTextFiles.WriteEOL(os.file);
  END Line;

(* ============================================================ *)

  PROCEDURE (os : IlasmFile)Locals(),NEW;
    VAR count : INTEGER;
	index : INTEGER;
	prcId : D.Scope;
	locId : D.Idnt;
  BEGIN
    count := 0;
   (* if dMax < 8, leave maxstack as default *)
    IF os.proc.dMax < 8 THEN os.CatStr("//") END;
    GPTextFiles.WriteChar(os.file, "	");
    os.CatStr(Asm.dirStr[Asm.dot_maxstack]); 
    os.Tint(os.proc.dMax);
    GPTextFiles.WriteEOL(os.file);

    GPTextFiles.WriteChar(os.file, "	");
    os.CatStr(Asm.dirStr[Asm.dot_locals]); 
    GPTextFiles.WriteChar(os.file, "(");
    IF os.proc.prId # NIL THEN 
      prcId := os.proc.prId;
      WITH prcId : Id.Procs DO
        FOR index := 0 TO prcId.locals.tide-1 DO
	  locId := prcId.locals.a[index];
	  IF ~(locId IS Id.ParId) THEN
	    IF count > 0 THEN os.Separator(',', 2) END;
	    os.TypeTag(locId.type);
	    os.Tidnt(locId);
	    INC(count);
	  END;
        END;
      ELSE (* nothing for module blocks *)
      END;
    END;
    WHILE count < os.proc.tLst.tide DO 
      IF count > 0 THEN os.Separator(',', 2) END;
      os.TypeTag(os.proc.tLst.a[count]);
      INC(count);
    END;
    GPTextFiles.WriteChar(os.file, ")");
    GPTextFiles.WriteEOL(os.file);
  END Locals;

(* ============================================================ *)
(*			Namehandling Methods			*)
(* ============================================================ *)

  PROCEDURE MkBlkName*(mod : Id.BlkId);
    VAR mNm : LitValue.CharOpen;
  (* --------------------------------------------------	*)
    PROCEDURE scpMangle(inS : LitValue.CharOpen) : LitValue.CharOpen;
      VAR outS : LitValue.CharOpen;
    BEGIN
      IF inS[0] = '[' THEN 
        outS := inS;
      ELSE
        outS := cat4(lBrk,inS,rBrk,inS);
      END;
      IF outS[LEN(outS)-1] # ']' THEN outS := cat2(outS, dotS) END;
      RETURN outS;
    END scpMangle;
  (* --------------------------------------------------	*)
    PROCEDURE nmSpaceOf(inS : LitValue.CharOpen) : LitValue.CharOpen;
      VAR ix : INTEGER;
	  ln : INTEGER;
	  ch : CHAR;
    BEGIN
      IF inS[0] # '[' THEN 
        RETURN inS;
      ELSE
	ln := LEN(inS);
	ix := 0;
        REPEAT
	  ch := inS[ix]; 
	  INC(ix);
	UNTIL (ix = LEN(inS)) OR (ch = ']');
        RETURN LitValue.subChOToChO(inS, ix, ln-ix);
      END;
    END nmSpaceOf;
  (* --------------------------------------------------	*)
  BEGIN
    IF mod.xName # NIL THEN RETURN END;
    IF mod.scopeNm # NIL THEN 
      mod.xName   := nmSpaceOf(mod.scopeNm);
      mod.scopeNm := scpMangle(mod.scopeNm); 
    ELSE
      mNm := D.getName.ChPtr(mod);
      mod.xName   := mNm;
     (*
      *  In the normal case, the assembly name is the
      *  same as the module name.  However, system 
      *  modules always have the assembly name "RTS".
      *)
      IF D.rtsMd IN mod.xAttr THEN 
	mod.scopeNm := cat3(lBrk, rtsS, rBrk);
      ELSE
	IF mod.kind = Id.impId THEN mNm := cat4(lBrk, mNm, rBrk, mNm) END;
        mod.scopeNm := cat2(mNm, dotS);
      END;
    END;
  END MkBlkName;

(* ------------------------------------------------------------ *)

  PROCEDURE MkRecName*(typ : Ty.Record);
    VAR mNm : LitValue.CharOpen;	(* prefix scope name    *)
	rNm : LitValue.CharOpen;	(* simple name of type  *)
	tId : D.Idnt;
	scp : D.Scope;
  (* ----------------------------------	*
   * The choice below may need revision	*
   * depending on any decison about the	*
   * format of foreign type-names	*
   * extracted from the metadata.	*
   * ----------------------------------	*)
    PROCEDURE unmangle(arr : LitValue.CharOpen) : LitValue.CharOpen;
    BEGIN
      RETURN arr;
    END unmangle;
  (* ----------------------------------------------------------	*)
  BEGIN
    IF typ.xName # NIL THEN RETURN END;

    IF (typ.baseTp IS Ty.Record) &
       (typ.baseTp.xName = NIL) THEN MkRecName(typ.baseTp(Ty.Record)) END;

    IF typ.extrnNm # NIL THEN
      typ.scopeNm := unmangle(typ.extrnNm);
    END;
    IF typ.bindTp # NIL THEN		(* Synthetically named rec'd *)
      tId := typ.bindTp.idnt;
      rNm := D.getName.ChPtr(tId);
    ELSE				(* Normal, named record type *)
      IF typ.idnt = NIL THEN		(* Anonymous record type     *)
        typ.idnt := Id.newAnonId(typ.serial);
	typ.idnt.type := typ;
      END;
      tId := typ.idnt;
      rNm := D.getName.ChPtr(tId);
    END;
    IF tId.dfScp = NIL THEN tId.dfScp := CompState.thisMod END;
    scp := tId.dfScp;
   (*
    *  At this program point the situation is at follows:
    *  rNm holds the simple name of the record. The scope
    *  in which the record is defined is scp.
    *)
    WITH scp : Id.Procs DO
	IF scp.prcNm = NIL THEN MkProcName(scp) END;
        rNm         := cat3(scp.prcNm, atSg, rNm);
        typ.xName   := rNm;
	typ.scopeNm := cat2(scp.scopeNm, rNm);
    | scp : Id.BlkId DO
	IF scp.xName = NIL THEN MkBlkName(scp) END;
        typ.xName   := rNm;
	typ.scopeNm := cat2(scp.scopeNm, rNm);
    END;
   (*
    *   It is at this point that we link records into the
    *   class-emission worklist.
    *)
    IF tId.dfScp.kind # Id.impId THEN
      LightningBase.emitter.AddNewRecEmitter(typ);
    END;
  END MkRecName;

(* ------------------------------------------------------------ *)

  PROCEDURE MkEnumName*(typ : Ty.Enum);
    VAR mNm : LitValue.CharOpen;	(* prefix scope name    *)
	rNm : LitValue.CharOpen;	(* simple name of type  *)
	tId : D.Idnt;
	scp : D.Scope;
  (* ----------------------------------------------------------	*)
  BEGIN
   (* Assert: Enums are always imported ... *)
    IF typ.xName # NIL THEN RETURN END;

    tId := typ.idnt;
    rNm := D.getName.ChPtr(tId);
    scp := tId.dfScp;
   (*
    *  At this program point the situation is at follows:
    *  rNm holds the simple name of the type. The scope
    *  in which the record is defined is scp.
    *)
    WITH  scp : Id.BlkId DO
      IF scp.xName = NIL THEN MkBlkName(scp) END;
      typ.xName := cat2(scp.scopeNm, rNm);
    END;
  END MkEnumName;

(* ------------------------------------------------------------ *)

  PROCEDURE MkPtrName*(typ : Ty.Pointer);
    VAR bndTp : D.Type;
	bndNm : LitValue.CharOpen;
  BEGIN
    bndTp := typ.boundTp;
    bndNm := typeName(bndTp);
    IF isValRecord(bndTp) THEN
      typ.xName := boxedName(bndTp(Ty.Record));
    ELSE
      typ.xName := bndNm;
    END;
  END MkPtrName;

(* ------------------------------------------------------------ *)

  PROCEDURE MkProcName*(proc : Id.Procs);
    VAR pNm : LitValue.CharOpen;
	res : Id.Procs;
	scp : Id.BlkId;
	bTp : Ty.Record;
  (* --------------------------------------------------	*)
    PROCEDURE MkMthNm(mth : Id.MthId);
      VAR res : Id.MthId;
	  scp : Id.BlkId;
	  typ : D.Type;
    BEGIN
      IF mth.scopeNm # NIL THEN RETURN;
      ELSIF mth.kind = Id.fwdMth THEN
	res := mth.resolve(Id.MthId); MkMthNm(res);
        mth.prcNm := res.prcNm; mth.scopeNm := res.scopeNm;
      ELSE
	scp := mth.dfScp(Id.BlkId);
	typ := mth.bndType;
	IF typ.xName = NIL THEN MkRecName(typ(Ty.Record)) END;
	IF scp.xName = NIL THEN MkBlkName(scp) END;
        mth.scopeNm := scp.scopeNm;
        IF mth.prcNm = NIL THEN mth.prcNm := D.getName.ChPtr(mth) END;
      END;
    END MkMthNm;
  (* --------------------------------------------------	*)
    PROCEDURE className(p : Id.Procs) : LitValue.CharOpen;
    BEGIN
      WITH p : Id.PrcId DO RETURN p.clsNm;
      |    p : Id.MthId DO RETURN p.bndType.xName;
      END;
    END className;
  (* --------------------------------------------------	*)
    PROCEDURE GetClassName(pr : Id.PrcId; bl : Id.BlkId);
      VAR nm : LitValue.CharOpen;
    BEGIN
      nm := D.getName.ChPtr(pr);
      IF pr.bndType = NIL THEN	(* normal procedure *)
	pr.clsNm := bl.xName;
        IF pr.prcNm = NIL THEN pr.prcNm := nm END;
      ELSE			(* static method *)
        IF pr.bndType.xName = NIL THEN MkRecName(pr.bndType(Ty.Record)) END;
	pr.clsNm := pr.bndType.xName; 
        IF pr.prcNm = NIL THEN 
	  pr.prcNm := nm;
	ELSIF pr.prcNm^ = initString THEN 
	  pr.SetKind(Id.ctorP);
        END;
      END;
    END GetClassName;
  (* --------------------------------------------------	*)
    PROCEDURE MkPrcNm(prc : Id.PrcId);
      VAR res : Id.PrcId;
	  scp : D.Scope;
    BEGIN
      IF prc.scopeNm # NIL THEN RETURN;
      ELSIF prc.kind = Id.fwdPrc THEN
	res := prc.resolve(Id.PrcId); MkPrcNm(res);
        prc.prcNm := res.prcNm; 
	prc.clsNm := res.clsNm;
	prc.scopeNm := res.scopeNm;
      ELSE
	scp := prc.dfScp;
	WITH scp : Id.BlkId DO
	    IF scp.xName = NIL THEN MkBlkName(scp) END;
	    IF  (D.isJv IN scp.xAttr) OR
		(D.isCs IN scp.xAttr) THEN
	      GetClassName(prc, scp);
	    ELSE
	      prc.clsNm := scp.xName;
              IF prc.prcNm = NIL THEN prc.prcNm := D.getName.ChPtr(prc) END;
	    END;
	| scp : Id.Procs DO
	    MkProcName(scp);
	    prc.clsNm := className(scp);
	    prc.prcNm   := cat3(D.getName.ChPtr(prc), atSg, scp.prcNm);
	END;
	prc.scopeNm := scp.scopeNm;
      END;
    END MkPrcNm;
  (* --------------------------------------------------	*)
  BEGIN
    WITH proc : Id.MthId DO MkMthNm(proc);
    |    proc : Id.PrcId DO MkPrcNm(proc);
    END;
  END MkProcName;

(* ------------------------------------------------------------ *)

  PROCEDURE MkAliasName*(typ : Ty.Opaque);
    VAR tNm : LitValue.CharOpen;
	sNm : LitValue.CharOpen;
  BEGIN
    IF typ.xName # NIL THEN RETURN END;
    MkBlkName(typ.idnt.dfScp(Id.BlkId));
    tNm := D.getName.ChPtr(typ.idnt);
    sNm := typ.idnt.dfScp.scopeNm;
    typ.xName   := cat2(sNm, tNm);
    typ.scopeNm := sNm;
  END MkAliasName;

(* ------------------------------------------------------------ *)

  PROCEDURE MkVarName*(var : Id.VarId);
  BEGIN
    IF var.varNm # NIL THEN RETURN END;
    var.varNm := D.getName.ChPtr(var);
    IF var.recTyp = NIL THEN	(* normal case *)
      var.clsNm := var.dfScp(Id.BlkId).xName;
    ELSE			(* static field *)
      var.clsNm := var.recTyp.xName; 
    END;
  END MkVarName;

(* ------------------------------------------------------------ *)

  PROCEDURE MkIdName*(id : D.Idnt);
  BEGIN
    WITH id : Id.Procs DO IF id.scopeNm = NIL THEN MkProcName(id) END;
    |    id : Id.BlkId DO IF id.scopeNm = NIL THEN MkBlkName(id) END;
    |    id : Id.VarId DO IF id.varNm   = NIL THEN MkVarName(id) END;
    |    id : Id.FldId DO IF id.fldNm   = NIL THEN 
				id.fldNm := D.getName.ChPtr(id) END;
    END;
  END MkIdName;

(* ------------------------------------------------------------ *)

  PROCEDURE NumberParams(pIdn : Id.Procs; pTyp : Ty.Procedure);
    VAR parId : Id.ParId;
	index : INTEGER;
	count : INTEGER;
   (* ----------------------------------------- *)
    PROCEDURE AppendTypeName(VAR lst : LitValue.CharOpenSeq; 
				 typ : D.Type;
				 idx : INTEGER);
    BEGIN
      IF idx > 0 THEN LitValue.AppendCharOpen(lst, cmma) END;
     (*
      *   We append type names, which must be lexically
      *   equivalent to the names used in the declaration
      *   in MethodDecl.
      *)
      IF typ.xName = NIL THEN MkTypeName(typ) END;
      WITH typ : Ty.Base DO
	  LitValue.AppendCharOpen(lst, typ.xName);
      | typ : Ty.Array DO
	  AppendTypeName(lst, typ.elemTp, 0);
	  LitValue.AppendCharOpen(lst, brks);
      | typ : Ty.Record DO
	  IF ~(D.clsTp IN typ.xAttr) THEN LitValue.AppendCharOpen(lst,vals) END;
	  LitValue.AppendCharOpen(lst, clss);
	  LitValue.AppendCharOpen(lst, typ.scopeNm);
      | typ : Ty.Pointer DO
	 (* 
	  *  This is a pointer to a value class, which has a
	  *  runtime representation as a boxed-class reference.
	  *)
	  IF isValRecord(typ.boundTp) THEN
	    LitValue.AppendCharOpen(lst, clss);
	    LitValue.AppendCharOpen(lst, typ.xName);
	  ELSE
	    AppendTypeName(lst, typ.boundTp,0);
	  END;
      | typ : Ty.Opaque DO
	  LitValue.AppendCharOpen(lst, clss);
	  LitValue.AppendCharOpen(lst, typ.xName);
      | typ : Ty.Enum DO
	  LitValue.AppendCharOpen(lst, vals);
	  LitValue.AppendCharOpen(lst, clss);
	  LitValue.AppendCharOpen(lst, typ.xName);
      END;
    END AppendTypeName;
   (* ----------------------------------------- *)
  BEGIN
    count := pTyp.argN;
    LitValue.ResetCharOpenSeq(fmArray);
    LitValue.AppendCharOpen(fmArray, lPar);
    FOR index := 0 TO pTyp.formals.tide-1 DO
      parId := pTyp.formals.a[index];
      parId.varOrd := count; INC(count);
     (* ilasm does not seem to need "[out]" here *)
      AppendTypeName(fmArray, parId.type, index);
      IF takeAdrs(parId) THEN 
	parId.boxOrd := parId.parMod;
	LitValue.AppendCharOpen(fmArray, rfMk);
      END; (* just mark *)
    END;
    LitValue.AppendCharOpen(fmArray, rPar);
    pTyp.xName := LitValue.arrayCat(fmArray);
   (*
    *   The current info.lNum (before the locals
    *   have been added) is the argsize.
    *)
    pTyp.argN := count;
    IF pTyp.retType # NIL THEN pTyp.retN := 1 END;
  END NumberParams;
 
(* ------------------------------------------------------------ *)

  PROCEDURE NumberLocals(pIdn : Id.Procs; IN locs : D.IdSeq);
    VAR ident : D.Idnt;
	index : INTEGER;
	count : INTEGER;
  BEGIN
    count := 0;
    FOR index := 0 TO locs.tide-1 DO
      ident := locs.a[index];
      WITH ident : Id.ParId DO (* skip *)
      | ident : Id.LocId DO
	  ident.varOrd := count;
	  INC(count);
      END;
    END;
    pIdn.rtsFram := count;
  END NumberLocals;

(* ------------------------------------------------------------ *)

  PROCEDURE MkCallAttr*(pIdn : D.Idnt); 
    VAR pTyp : Ty.Procedure;
	rcvP : Id.ParId;
  BEGIN
    pTyp := pIdn.type(Ty.Procedure);
    WITH pIdn : Id.MthId DO
	pTyp.argN := 1;		(* count one for "this" *)
	rcvP := pIdn.rcvFrm;
(*
THIS NEED CORRECTING FOR VAR, POINTER RECEIVERS
	IF ~takeAdrs(pIdn.rcvFrm) THEN 
	  pTyp.argN := 1;	(* count one for "this" *)
	ELSE
	  pTyp.argN := 2;	(* this plus the retbox *)
	END;
 *)
	MkProcName(pIdn);
        IF takeAdrs(rcvP) THEN rcvP.boxOrd := rcvP.parMod END;
	NumberParams(pIdn, pTyp);
    | pIdn : Id.PrcId DO 
	pTyp.argN := 0;
	MkProcName(pIdn);
	NumberParams(pIdn, pTyp);
    END;
  END MkCallAttr;

(* ------------------------------------------------------------ *)

  PROCEDURE RenumberLocals*(prcId : Id.Procs);
    VAR parId : Id.ParId;
	frmTp : Ty.Procedure;
	funcT : BOOLEAN;
  BEGIN
   (*
    * Rules:
    * 	(i)   The receiver (if any) must be #0
    * 	(ii)  Params are #0 .. #N for statics, or #1 .. #N
    *         for methods, or #2 .. #N for methods with a VAR receiver.
    * 	(iii) Locals separately number from zero.
    * 
    *   This procedure computes the number of local slots. It
    *   renumbers the varOrd fields, and initializes rtsFram.
    *   The procedure also computes the formal name for the JVM.
    *) 
    frmTp := prcId.type(Ty.Procedure);
    funcT := (frmTp.retType # NIL);
    WITH prcId : Id.MthId DO
      parId := prcId.rcvFrm;
      parId.varOrd := 0;   
      IF takeAdrs(parId) THEN parId.boxOrd := parId.parMod END;
      frmTp.argN := 1;		(* count one for "this" *)
    ELSE (* static procedures *)
      frmTp.argN := 0;	
    END;
   (*
    *   Assert: params do not appear in the local array.
    *   Count params (and boxes if needed).
    *)
    NumberParams(prcId, frmTp);
    IF prcId.body # NIL THEN
      NumberLocals(prcId, prcId.locals);
    END;
  END RenumberLocals;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

  PROCEDURE (os : IlasmFile)LdIndirect*(typ : D.Type),NEW;
    VAR code : INTEGER;
  BEGIN
    IF (typ # NIL) & (typ IS Ty.Base) THEN
      os.Code(typeLdInd[typ(Ty.Base).tpOrd]);
    ELSIF isValRecord(typ) THEN
      os.CodeT(Asm.opc_ldobj, typ);
    ELSE
      os.Code(Asm.opc_ldind_ref);
    END;
  END LdIndirect;

(* ------------------------------------------------------------ *)

  PROCEDURE (os : IlasmFile)StIndirect*(typ : D.Type),NEW;
    VAR code : INTEGER;
  BEGIN
    IF (typ # NIL) & (typ IS Ty.Base) THEN
      os.Code(typeStInd[typ(Ty.Base).tpOrd]);
    ELSIF isValRecord(typ) THEN
      os.CodeT(Asm.opc_stobj, typ);
    ELSE
      os.Code(Asm.opc_stind_ref);
    END;
  END StIndirect;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

  PROCEDURE (os : IlasmFile)LoadArg*(ord : INTEGER),NEW;
  BEGIN
    IF ord < 256 THEN
      CASE ord OF
      | 0 : os.Code(Asm.opc_ldarg_0);
      | 1 : os.Code(Asm.opc_ldarg_1);
      | 2 : os.Code(Asm.opc_ldarg_2);
      | 3 : os.Code(Asm.opc_ldarg_3);
      ELSE
	os.CodeI(Asm.opc_ldarg_s, ord);
      END;
    ELSE
      os.CodeI(Asm.opc_ldarg, ord);
    END;
  END LoadArg;

(* ------------------------------------------------------------ *)

  PROCEDURE (os : IlasmFile)LoadLocal*(ord : INTEGER),NEW;
  BEGIN
    IF ord < 256 THEN
      CASE ord OF
      | 0 : os.Code(Asm.opc_ldloc_0);
      | 1 : os.Code(Asm.opc_ldloc_1);
      | 2 : os.Code(Asm.opc_ldloc_2);
      | 3 : os.Code(Asm.opc_ldloc_3);
      ELSE
	os.CodeI(Asm.opc_ldloc_s, ord);
      END;
    ELSE
      os.CodeI(Asm.opc_ldloc, ord);
    END;
  END LoadLocal;

(* ----------------------------------------------------	*)

  PROCEDURE (os : IlasmFile)LoadLocalA*(ord : INTEGER),NEW;
  BEGIN
    IF ord < 256 THEN
      os.CodeI(Asm.opc_ldloca_s, ord);
    ELSE
      os.CodeI(Asm.opc_ldloca, ord);
    END;
  END LoadLocalA;

(* ---------------------------------------------------- *)

  PROCEDURE (os : IlasmFile)GetLocal*(var : Id.LocId),NEW;
  BEGIN
    WITH var : Id.ParId DO
      os.LoadArg(var.varOrd);
      IF var.boxOrd # D.val THEN os.LdIndirect(var.type) END;
    ELSE
      os.LoadLocal(var.varOrd);
    END;
  END GetLocal;

(* ---------------------------------------------------- *)

  PROCEDURE (os : IlasmFile)DecTemp*(ord : INTEGER),NEW;
  BEGIN
    os.LoadLocal(ord);
    os.Code(Asm.opc_ldc_i4_1);
    os.Code(Asm.opc_sub);
    os.StoreLocal(ord);
  END DecTemp;

(* ---------------------------------------------------- *)

  PROCEDURE (os : IlasmFile)GetVar*(id : D.Idnt),NEW;
    VAR var : Id.AbVar;
	scp : D.Scope;
  BEGIN
    var := id(Id.AbVar);
    IF var.kind = Id.conId THEN
      os.GetLocal(var(Id.LocId));
    ELSE
      scp := var.dfScp;
      WITH scp : Id.BlkId DO
        os.PutGetS(Asm.opc_ldsfld, scp, var(Id.VarId));
      ELSE
        os.GetLocal(var(Id.LocId));
      END;
    END;
  END GetVar;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

  PROCEDURE (os : IlasmFile)GetLocalA(var : Id.LocId),NEW;
    VAR ord : INTEGER;
  BEGIN
    ord := var.varOrd;
    WITH var : Id.ParId DO
      IF var.boxOrd # D.val THEN 
        (* nothing to do, handle already loaded *)
      ELSIF ord < 256 THEN
	os.CodeI(Asm.opc_ldarga_s, ord);
      ELSE
	os.CodeI(Asm.opc_ldarga, ord);
      END;
    ELSE
      os.LoadLocalA(ord);
    END;
  END GetLocalA;

(* ---------------------------------------------------- *)

  PROCEDURE (os : IlasmFile)GetVarA*(id : D.Idnt),NEW;
    VAR var : Id.AbVar;
	scp : D.Scope;
  BEGIN
   (*
    *  Assert:  the handle is already pushed on the tos.
    *)
    var := id(Id.AbVar);
    scp := var.dfScp;
    WITH scp : Id.BlkId DO
      os.PutGetS(Asm.opc_ldsflda, scp, var(Id.VarId));
    ELSE
      os.GetLocalA(var(Id.LocId));
    END;
  END GetVarA;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

  PROCEDURE (os : IlasmFile)StoreArg*(ord : INTEGER),NEW;
  BEGIN
    IF ord < 256 THEN
      os.CodeI(Asm.opc_starg_s, ord);
    ELSE
      os.CodeI(Asm.opc_starg, ord);
    END;
  END StoreArg;

(* ---------------------------------------------------- *)

  PROCEDURE (os : IlasmFile)StoreLocal*(ord : INTEGER),NEW;
  BEGIN
    IF ord < 256 THEN
      CASE ord OF
      | 0 : os.Code(Asm.opc_stloc_0);
      | 1 : os.Code(Asm.opc_stloc_1);
      | 2 : os.Code(Asm.opc_stloc_2);
      | 3 : os.Code(Asm.opc_stloc_3);
      ELSE
	os.CodeI(Asm.opc_stloc_s, ord);
      END;
    ELSE
      os.CodeI(Asm.opc_stloc, ord);
    END;
  END StoreLocal;

(* ---------------------------------------------------- *)

  PROCEDURE (os : IlasmFile)PutLocal*(var : Id.LocId),NEW;
  BEGIN
    WITH var : Id.ParId DO
      IF var.boxOrd = D.val THEN 
	os.StoreArg(var.varOrd);
      ELSE
       (*
	*  stack goes (top) value, reference, ... so
	*  os.LoadArg(var.varOrd);
	*)
	os.StIndirect(var.type);
      END;
    ELSE
      os.StoreLocal(var.varOrd);
    END;
  END PutLocal;

(* ---------------------------------------------------- *)

  PROCEDURE (os : IlasmFile)PutVar*(id : D.Idnt),NEW;
    VAR var : Id.AbVar;
	scp : D.Scope;
  BEGIN
    var := id(Id.AbVar);
    scp := var.dfScp;
    WITH scp : Id.BlkId DO
      os.PutGetS(Asm.opc_stsfld, scp, var(Id.VarId));
    ELSE (* must be local *)
      os.PutLocal(var(Id.LocId));
    END;
  END PutVar;

(* ------------------------------------------------------------ *)

  PROCEDURE (os : IlasmFile)PutElem*(typ : D.Type),NEW;
  (* typ is element type *)
  BEGIN
    IF (typ # NIL) & (typ IS Ty.Base) THEN
      os.Code(typePutE[typ(Ty.Base).tpOrd]);
    ELSIF isValRecord(typ) THEN
      os.CodeT(Asm.opc_stobj, typ);
    ELSE
      os.Code(Asm.opc_stelem_ref);
    END;
  END PutElem;

(* ------------------------------------------------------------ *)

  PROCEDURE (os : IlasmFile)GetElem*(typ : D.Type),NEW;
  BEGIN
    IF (typ # NIL) & (typ IS Ty.Base) THEN
      os.Code(typeGetE[typ(Ty.Base).tpOrd]);
    ELSIF isValRecord(typ) THEN
      os.CodeT(Asm.opc_ldobj, typ);
    ELSE
      os.Code(Asm.opc_ldelem_ref);
    END;
  END GetElem;

(* ------------------------------------------------------------ *)

  PROCEDURE (os : IlasmFile)GetElemA*(typ : D.Type),NEW;
  BEGIN
    os.CodeTn(Asm.opc_ldelema, typ);
  END GetElemA;

(* ------------------------------------------------------------ *)

  PROCEDURE (os : IlasmFile)GetVal*(typ : Ty.Pointer),NEW;
  BEGIN
    os.PutGetFld(Asm.opc_ldfld, typ.boundTp, typ, vFld);
  END GetVal;

(* ------------------------------------------------------------ *)

  PROCEDURE (os : IlasmFile)GetValA*(typ : Ty.Pointer),NEW;
  BEGIN
    os.PutGetFld(Asm.opc_ldflda, typ.boundTp, typ, vFld);
  END GetValA;

(* ------------------------------------------------------------ *)

  PROCEDURE (os : IlasmFile)PushInt*(num : INTEGER),NEW;
  BEGIN
    IF (-128 <= num) & (num <= 127) THEN
      CASE num OF
      | -1 : os.Code(Asm.opc_ldc_i4_M1);
      |  0 : os.Code(Asm.opc_ldc_i4_0);
      |  1 : os.Code(Asm.opc_ldc_i4_1);
      |  2 : os.Code(Asm.opc_ldc_i4_2);
      |  3 : os.Code(Asm.opc_ldc_i4_3);
      |  4 : os.Code(Asm.opc_ldc_i4_4);
      |  5 : os.Code(Asm.opc_ldc_i4_5);
      |  6 : os.Code(Asm.opc_ldc_i4_6);
      |  7 : os.Code(Asm.opc_ldc_i4_7);
      |  8 : os.Code(Asm.opc_ldc_i4_8);
      ELSE
        os.CodeI(Asm.opc_ldc_i4_s, num);
      END;
    ELSE
      os.CodeI(Asm.opc_ldc_i4, num);
    END;
  END PushInt;

(* ------------------------------------------------------------ *)

  PROCEDURE (os : IlasmFile)PushLong*(num : LONGINT),NEW;
  BEGIN
   (*
    *  IF num is short we could do PushInt, then i2l!
    *)
    os.CodeL(Asm.opc_ldc_i8, num);
  END PushLong;

(* ------------------------------------------------------------ *)

  PROCEDURE (os : IlasmFile)PushReal*(num : REAL),NEW;
  BEGIN
    os.CodeR(Asm.opc_ldc_r8, num);
  END PushReal;

(* ------------------------------------------------------------ *)

  PROCEDURE (os : IlasmFile)PushSReal*(num : REAL),NEW;
  BEGIN
    os.CodeR(Asm.opc_ldc_r4, num);
  END PushSReal;

(* ------------------------------------------------------------ *)

  PROCEDURE (os : IlasmFile)PushJunkAndQuit*(prc : D.Scope),NEW;
    VAR pTyp : Ty.Procedure;
  BEGIN
    IF (prc # NIL) & (prc.type # NIL) THEN
      pTyp := prc.type(Ty.Procedure);
      IF pTyp.retType # NIL THEN os.PushZero(pTyp.retType) END;
    END;
    os.DoReturn();
  END PushJunkAndQuit;

(* ------------------------------------------------------------ *)

  PROCEDURE (os : IlasmFile)ConvertUp*(inT, outT : D.Type),NEW;
   (* Conversion "up" is always safe at runtime. Many are nop.	*)
    VAR inB, outB, code : INTEGER;
  BEGIN
    inB  := inT(Ty.Base).tpOrd;
    outB := outT(Ty.Base).tpOrd;
    IF inB = outB THEN RETURN END;		(* PREMATURE RETURN! *)
    CASE outB OF
    | Ty.realN : code := Asm.opc_conv_r8;
    | Ty.sReaN : code := Asm.opc_conv_r4;
    | Ty.lIntN : code := Asm.opc_conv_i8;
    ELSE	 	RETURN; 		(* PREMATURE RETURN! *)
    END;
    os.Code(code);
  END ConvertUp;

(* ------------------------------------------------------------ *)

  PROCEDURE (os : IlasmFile)ConvertDn*(inT, outT : D.Type),NEW;
   (* Conversion "down" often needs a runtime check. *)
    VAR inB, outB, code : INTEGER;
  BEGIN
    inB  := inT(Ty.Base).tpOrd;
    outB := outT(Ty.Base).tpOrd;
    IF inB = Ty.setN THEN inB := Ty.intN END;
    IF inB = outB THEN RETURN END;		(* PREMATURE RETURN! *)
    IF os.proc.prId.ovfChk THEN
      CASE outB OF
      | Ty.realN : RETURN;			(* PREMATURE RETURN! *)
      | Ty.sReaN : code := Asm.opc_conv_r4;	(* No check possible *)
      | Ty.lIntN : code := Asm.opc_conv_ovf_i8;
      | Ty.intN  : code := Asm.opc_conv_ovf_i4;
      | Ty.sIntN : code := Asm.opc_conv_ovf_i2;
      | Ty.byteN : code := Asm.opc_conv_ovf_i1;
      | Ty.setN  : code := Asm.opc_conv_u4;	(* no check here! *)
      | Ty.charN : code := Asm.opc_conv_ovf_u2;
      | Ty.sChrN : code := Asm.opc_conv_ovf_u1;
      END;
    ELSE
      CASE outB OF
      | Ty.realN : RETURN;			(* PREMATURE RETURN! *)
      | Ty.sReaN : code := Asm.opc_conv_r4;	(* No check possible *)
      | Ty.lIntN : code := Asm.opc_conv_i8;
      | Ty.intN  : code := Asm.opc_conv_i4;
      | Ty.sIntN : code := Asm.opc_conv_i2;
      | Ty.byteN : code := Asm.opc_conv_i1;
      | Ty.setN  : code := Asm.opc_conv_u4;	(* no check here! *)
      | Ty.charN : code := Asm.opc_conv_u2;
      | Ty.sChrN : code := Asm.opc_conv_u1;
      END;
    END;
    os.Code(code);
  END ConvertDn;

(* ------------------------------------------------------------ *)

  PROCEDURE (os : IlasmFile)EmitOneRange*
	       (var : INTEGER;		(* local variable index *)
		loC : INTEGER;		(* low-value of range   *)
		hiC : INTEGER;		(* high-value of range  *)
		ord : INTEGER;		(* case-index of range  *)
		min : INTEGER;		(* minimun selector val *)
		max : INTEGER;		(* maximum selector val *)
		def : INTEGER),NEW;	(* default code label   *)
  (* ---------------------------------------------------------- *
   *  The selector value is known to be in the range min .. max *
   *  and we wish to send values between loC and hiC to the     *
   *  code label associated with ord. All otherwise go to def.  *
   *  A range is "compact" if it is hard against min/max limits *
   * ---------------------------------------------------------- *)
    VAR target : INTEGER;
  BEGIN
   (*
    *    Deal with several special cases...
    *)
    target := def + ord + 1;
    IF (min = loC) & (max = hiC) THEN	(* fully compact: just GOTO *)
      os.CodeLb(Asm.opc_br, target);
    ELSE
      os.LoadLocal(var);
      IF loC = hiC THEN (* a singleton *)
	os.PushInt(loC);
	os.CodeLb(Asm.opc_beq, target);
      ELSIF min = loC THEN		(* compact at low end only  *)
	os.PushInt(hiC);
	os.CodeLb(Asm.opc_ble, target);
      ELSIF max = hiC THEN		(* compact at high end only *)
	os.PushInt(loC);
	os.CodeLb(Asm.opc_bge, target);
      ELSE				(* Shucks! The general case *)
	os.PushInt(loC);
	os.CodeLb(Asm.opc_blt, def);
	os.LoadLocal(var);
	os.PushInt(hiC);
	os.CodeLb(Asm.opc_ble, target);
      END;
      os.CodeLb(Asm.opc_br, def);
    END;
  END EmitOneRange;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

  PROCEDURE (os : IlasmFile)InitVars*(scp : D.Scope),NEW;
    VAR index : INTEGER;
	ident : D.Idnt;
  BEGIN
   (* 
    *  Initialize local fields, if needed
    *)
    FOR index := 0 TO scp.locals.tide-1 DO
      ident := scp.locals.a[index];
      IF ~(ident IS Id.ParId) THEN
        IF ~ident.type.isScalarType() THEN 
	  os.StructInit(ident);
	ELSE
	  WITH ident : Id.LocId DO
	   (*
	    *  Special code to step around deficiency in the the
	    *  verifier. Verifier does not understand OUT semantics.
	    *)
	    IF ident(Id.LocId).addrsd THEN
	      ASSERT(~(scp IS Id.BlkId));
	      os.ScalarInit(ident); 
	      os.StoreLocal(ident.varOrd);
	    END;
	  ELSE
	  END;
	END;
      END;
    END;
  END InitVars;

(* ============================================================ *)
(* ============================================================ *)
BEGIN
  LitValue.InitCharOpenSeq(fmArray, 8);
  LitValue.InitCharOpenSeq(nmArray, 8);

  mtPr := LitValue.strToCharOpen("()"); 
  fpzS := LitValue.strToCharOpen("0.0"); 
  rtsS := LitValue.strToCharOpen("RTS"); 
  cln2 := LitValue.strToCharOpen("::"); 
  brks := LitValue.strToCharOpen("[]"); 
  dotS := LitValue.strToCharOpen("."); 
  cmma := LitValue.strToCharOpen(","); 
  semi := LitValue.strToCharOpen(";"); 
  lPar := LitValue.strToCharOpen("("); 
  rPar := LitValue.strToCharOpen(")"); 
  lBrk := LitValue.strToCharOpen("["); 
  rBrk := LitValue.strToCharOpen("]"); 
  lowL := LitValue.strToCharOpen("_"); 
  sQut := LitValue.strToCharOpen(","); 
  atSg := LitValue.strToCharOpen("@"); 
  rfMk := LitValue.strToCharOpen("&"); 
  vFld := LitValue.strToCharOpen("v$"); 
  ouMk := LitValue.strToCharOpen("[out] ");
  clss := LitValue.strToCharOpen("class "); 
  vals := LitValue.strToCharOpen("value "); 
  invalid := LitValue.strToCharOpen("<<don't care>>"); 
  boxedObj := LitValue.strToCharOpen("Boxed_"); 

  G.setTp.xName  := LitValue.strToCharOpen("int32");
  G.boolTp.xName := LitValue.strToCharOpen("bool");
  G.byteTp.xName := LitValue.strToCharOpen("int8");
  G.charTp.xName := LitValue.strToCharOpen("wchar");
  G.sChrTp.xName := LitValue.strToCharOpen("char");
  G.sIntTp.xName := LitValue.strToCharOpen("int16");
  G.lIntTp.xName := LitValue.strToCharOpen("int64");
  G.realTp.xName := LitValue.strToCharOpen("float64");
  G.sReaTp.xName := LitValue.strToCharOpen("float32");
  G.intTp.xName  := G.setTp.xName;
  G.anyRec.xName := LitValue.strToCharOpen("class System.Object");
  G.anyPtr.xName := G.anyRec.xName;

  typeGetE[ Ty.boolN] := Asm.opc_ldelem_i1;
  typeGetE[ Ty.sChrN] := Asm.opc_ldelem_u1;
  typeGetE[ Ty.charN] := Asm.opc_ldelem_u2;
  typeGetE[ Ty.byteN] := Asm.opc_ldelem_i1;
  typeGetE[ Ty.sIntN] := Asm.opc_ldelem_i2;
  typeGetE[  Ty.intN] := Asm.opc_ldelem_i4;
  typeGetE[ Ty.lIntN] := Asm.opc_ldelem_i8;
  typeGetE[ Ty.sReaN] := Asm.opc_ldelem_r4;
  typeGetE[ Ty.realN] := Asm.opc_ldelem_r8;
  typeGetE[  Ty.setN] := Asm.opc_ldelem_i4;
  typeGetE[Ty.anyPtr] := Asm.opc_ldelem_ref;
  typeGetE[Ty.anyRec] := Asm.opc_ldelem_ref;

  typePutE[ Ty.boolN] := Asm.opc_stelem_i1;
  typePutE[ Ty.sChrN] := Asm.opc_stelem_i1;
  typePutE[ Ty.charN] := Asm.opc_stelem_i2;
  typePutE[ Ty.byteN] := Asm.opc_stelem_i1;
  typePutE[ Ty.sIntN] := Asm.opc_stelem_i2;
  typePutE[  Ty.intN] := Asm.opc_stelem_i4;
  typePutE[ Ty.lIntN] := Asm.opc_stelem_i8;
  typePutE[ Ty.sReaN] := Asm.opc_stelem_r4;
  typePutE[ Ty.realN] := Asm.opc_stelem_r8;
  typePutE[  Ty.setN] := Asm.opc_stelem_i4;
  typePutE[Ty.anyPtr] := Asm.opc_stelem_ref;
  typePutE[Ty.anyRec] := Asm.opc_stelem_ref;

  typeLdInd[ Ty.boolN] := Asm.opc_ldind_u1;
  typeLdInd[ Ty.sChrN] := Asm.opc_ldind_u1;
  typeLdInd[ Ty.charN] := Asm.opc_ldind_u2;
  typeLdInd[ Ty.byteN] := Asm.opc_ldind_i1;
  typeLdInd[ Ty.sIntN] := Asm.opc_ldind_i2;
  typeLdInd[  Ty.intN] := Asm.opc_ldind_i4;
  typeLdInd[ Ty.lIntN] := Asm.opc_ldind_i8;
  typeLdInd[ Ty.sReaN] := Asm.opc_ldind_r4;
  typeLdInd[ Ty.realN] := Asm.opc_ldind_r8;
  typeLdInd[  Ty.setN] := Asm.opc_ldind_i4;
  typeLdInd[Ty.anyPtr] := Asm.opc_ldind_ref;
  typeLdInd[Ty.anyRec] := Asm.opc_ldind_ref;

  typeStInd[ Ty.boolN] := Asm.opc_stind_i1;
  typeStInd[ Ty.sChrN] := Asm.opc_stind_i1;
  typeStInd[ Ty.charN] := Asm.opc_stind_i2;
  typeStInd[ Ty.byteN] := Asm.opc_stind_i1;
  typeStInd[ Ty.sIntN] := Asm.opc_stind_i2;
  typeStInd[  Ty.intN] := Asm.opc_stind_i4;
  typeStInd[ Ty.lIntN] := Asm.opc_stind_i8;
  typeStInd[ Ty.sReaN] := Asm.opc_stind_r4;
  typeStInd[ Ty.realN] := Asm.opc_stind_r8;
  typeStInd[  Ty.setN] := Asm.opc_stind_i4;
  typeStInd[Ty.anyPtr] := Asm.opc_stind_ref;
  typeStInd[Ty.anyRec] := Asm.opc_stind_ref;

END IlasmUtil.
(* ============================================================ *)
(* ============================================================ *)

